"""Generate Java code for XML-ization based on the intermediate representation."""
import io
import textwrap

from typing import Tuple, Optional, List

from icontract import ensure, require

from aas_core_codegen import intermediate, naming, specific_implementations
from aas_core_codegen.common import (
    assert_never,
    Error,
    Identifier,
    indent_but_first_line,
    Stripped,
)
from aas_core_codegen.java import (
    common as java_common,
    naming as java_naming,
)
from aas_core_codegen.java.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
    INDENT6 as IIIIII,
)

# region Generate


def _generate_result() -> Stripped:
    """Generate the class to represent XML de/serialize results."""
    return Stripped(
        f"""\
private static class Result<T> {{
{I}private final T result;
{I}private final Reporting.Error error;
{I}private final boolean success;

{I}private Result(T result, Reporting.Error error, boolean success) {{
{II}this.result = result;
{II}this.error = error;
{II}this.success = success;
{I}}}

{I}public static <T> Result<T> success(T result) {{
{II}if(result == null) throw new IllegalArgumentException("Result must not be null.");
{II}return new Result<>(result, null, true);
{I}}}

{I}public static <T> Result<T> failure(Reporting.Error error) {{
{II}if(error == null) throw new IllegalArgumentException("Error must not be null.");
{II}return new Result<>(null, error, false);
{I}}}

{I}@SuppressWarnings("unchecked")
{I}public <I> Result<I> castTo(Class<I> type){{
{II}if(isError() || type.isInstance(result)) return (Result<I>) this;
{II}throw new IllegalStateException("Result of type "
{III}+ result.getClass().getName()
{III}+ " is not an instance of "
{III}+ type.getName());
{I}}}

{I}public T getResult() {{
{II}if (!isSuccess()) throw new IllegalStateException("Result is not present.");
{II}return result;
{I}}}

{I}public boolean isSuccess() {{
{II}return success;
{I}}}

{I}public boolean isError(){{return !success;}}

{I}public Reporting.Error getError() {{
{II}if (isSuccess()) throw new IllegalStateException("Result is present.");
{II}return error;
{I}}}

{I}public <R> R map(Function<T, R> successFunction, Function<Reporting.Error, R> errorFunction) {{
{II}return isSuccess() ? successFunction.apply(result) : errorFunction.apply(error);
{I}}}

{I}public T onError(Function<Reporting.Error, T>  errorFunction){{
{II}return map(Function.identity(), errorFunction);
{I}}}
}}"""
    )


def _generate_current_event() -> Stripped:
    """Generate the function to a single XML event."""

    return Stripped(
        f"""\
private static XMLEvent currentEvent(XMLEventReader reader) {{
{I}try {{
{II}return reader.peek();
{I}}} catch (XMLStreamException xmlStreamException) {{
{II}throw new Xmlization.DeserializeException("",
{III}"Failed in method peek because of: " +
{III}xmlStreamException.getMessage());
{I}}}
}}"""
    )


def _generate_try_content_for_primitives() -> Stripped:
    """Generate the function to read textual content."""

    return Stripped(
        f"""\
private static String readContentAsString(XMLEventReader reader) throws XMLStreamException {{
{I}final StringBuilder content = new StringBuilder();

{I}while (reader.peek().isCharacters() || reader.peek().getEventType() == XMLStreamConstants.COMMENT) {{
{II}if (reader.peek().isCharacters()) {{
{III}content.append(reader.peek().asCharacters().getData());
{II}}}
{II}reader.nextEvent();
{I}}}

{I}return content.toString();
}}

private static Boolean readContentAsBool(XMLEventReader reader) throws XMLStreamException {{
{I}final StringBuilder content = new StringBuilder();

{I}while (reader.peek().isCharacters() || reader.peek().getEventType() == XMLStreamConstants.COMMENT) {{
{II}if (reader.peek().isCharacters()) {{
{III}content.append(reader.peek().asCharacters().getData());
{II}}}
{II}reader.nextEvent();
{I}}}
{I}if(!("true".equals(content.toString()) || "false".equals(content.toString()))){{
{II}throw new IllegalStateException("Content cannot be converted to the type Boolean.");
{I}}}
{I}return Boolean.valueOf(content.toString());
}}

private static Long readContentAsLong(XMLEventReader reader) throws XMLStreamException {{
{I}final StringBuilder content = new StringBuilder();

{I}while (reader.peek().isCharacters() || reader.peek().getEventType() == XMLStreamConstants.COMMENT) {{
{II}if (reader.peek().isCharacters()) {{
{III}content.append(reader.peek().asCharacters().getData());
{II}}}
{II}reader.nextEvent();
{I}}}

{I}return Long.valueOf(content.toString());
}}

private static Double readContentAsDouble(XMLEventReader reader) throws XMLStreamException {{
{I}final StringBuilder content = new StringBuilder();

{I}while (reader.peek().isCharacters() || reader.peek().getEventType() == XMLStreamConstants.COMMENT) {{
{II}if (reader.peek().isCharacters()) {{
{III}content.append(reader.peek().asCharacters().getData());
{II}}}
{II}reader.nextEvent();
{I}}}

{I}return Double.valueOf(content.toString());
}}

/**
 * Read the whole content of an element into memory.
 */
private static byte[] readContentAsBase64(
{I}XMLEventReader reader) throws XMLStreamException {{
{I}final StringBuilder content = new StringBuilder();
{I}while (reader.peek().isCharacters() || reader.peek().getEventType() == XMLStreamConstants.COMMENT) {{
{II}if (reader.peek().isCharacters()) {{
{III}content.append(reader.peek().asCharacters().getData());
{II}}}
{II}reader.nextEvent();
{I}}}

{I}String encodedData = content.toString();
{I}final byte[] decodedData;
{I}Base64.Decoder decoder = Base64.getDecoder();

{I}try {{
{II}decodedData = decoder.decode(encodedData);
{I}}} catch (IllegalArgumentException exception) {{
{II}throw new XMLStreamException(
{III}"Failed to read base64 encoded data: " +
{III}exception.getMessage());
{I}}}

{I}return decodedData;
}}"""
    )


def _generate_skip_whitespace_and_comments() -> Stripped:
    """Generate the function to skip whitespace text and XML comments."""
    return Stripped(
        f"""\
private static void skipWhitespaceAndComments(XMLEventReader reader) {{
{I}while (whiteSpaceOrComment(reader)) {{
{II}reader.next();
{I}}}
}}

private static boolean whiteSpaceOrComment(XMLEventReader reader) {{
{I}final XMLEvent currentEvent = currentEvent(reader);
{I}final boolean isComment = (currentEvent != null &&
{II}currentEvent.getEventType() == XMLStreamConstants.COMMENT);
{I}final boolean isWhiteSpace = (currentEvent != null &&
{II}currentEvent.getEventType() == XMLStreamConstants.CHARACTERS &&
{II}currentEvent.asCharacters().isWhiteSpace());
{I}return isComment || isWhiteSpace;
}}"""
    )


def _generate_skip_start_document() -> Stripped:
    """Generate the function to skip start document."""
    return Stripped(
        f"""\
private static void skipStartDocument(XMLEventReader reader){{
{I}if (currentEvent(reader).isStartDocument()){{
{II}reader.next();
{I}}}
}}"""
    )


def _generate_is_empty_element() -> Stripped:
    """Generate the function to check if an element is empty."""
    return Stripped(
        f"""\
private static boolean isEmptyElement(XMLEventReader reader) {{
{I}// Skip the element node and go to the content
{I}try {{
{II}reader.nextEvent();
{I}}} catch (XMLStreamException xmlStreamException) {{
{II}throw new Xmlization.DeserializeException("",
{III}"Failed in method isEmptyElement because of: " +
{III}xmlStreamException.getMessage());
{I}}}
{I}return currentEvent(reader).isEndElement();
}}"""
    )


def _generate_deserialize_primitive_property(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> Stripped:
    """Generate the snippet to deserialize a property ``prop`` of primitive type."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    a_type = intermediate.try_primitive_type(type_anno)
    assert a_type is not None, f"Unexpected type annotation: {prop.type_annotation}"

    deserialization_expr: str
    if a_type is intermediate.PrimitiveType.BOOL:
        deserialization_expr = "readContentAsBool(reader)"
    elif a_type is intermediate.PrimitiveType.INT:
        deserialization_expr = "readContentAsInt(reader)"
    elif a_type is intermediate.PrimitiveType.FLOAT:
        deserialization_expr = "readContentAsFloat(reader)"
    elif a_type is intermediate.PrimitiveType.STR:
        deserialization_expr = "readContentAsString(reader)"
    elif a_type is intermediate.PrimitiveType.BYTEARRAY:
        deserialization_expr = "readContentAsBase64(reader)"
    else:
        assert_never(a_type)

    target_var = java_naming.variable_name(Identifier(f"the_{prop.name}"))

    prop_name = java_naming.property_name(prop.name)
    cls_name = java_naming.class_name(cls.name)
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    if a_type is intermediate.PrimitiveType.STR:
        empty_handling_body = Stripped(f'{target_var} = "";')
    else:
        empty_handling_body = Stripped(
            f"""\
final Reporting.Error error = new Reporting.Error(
{I}"The property {prop_name} of an instance of class {cls_name} " +
{I}"can not be de-serialized from a self-closing element " +
{I}"since it needs content");
error.prependSegment(
{I}new Reporting.NameSegment(
{II}{xml_prop_name_literal}));
return Result.failure(error);"""
        )

    return Stripped(
        f"""\
if (isEmptyProperty) {{
{I}{indent_but_first_line(empty_handling_body, I)}
}}
else {{
{I}if (currentEvent(reader).isEndDocument()) {{
{II}final Reporting.Error error = new Reporting.Error(
{III}"Expected an XML content representing " +
{III}"the property {prop_name} of an instance of class {cls_name}, " +
{III}"but reached the end-of-file");
{II}return Result.failure(error);
{I}}}

{I}try {{
{II}{target_var} = {deserialization_expr};
{I}}} catch (Exception e) {{
{II}final Reporting.Error error = new Reporting.Error(
{III}"The property {prop_name} of an instance of class {cls_name} "
{IIII}+ " could not be de-serialized: " + e.getMessage());
{II}error.prependSegment(
{III}new Reporting.NameSegment(
{IIII}"{prop_name}"));
{II}return Result.failure(error);
{I}}}
}}"""
    )


def _generate_get_event_type_as_string() -> Stripped:
    """Generate the function to map XML event types to their string representations."""

    return Stripped(
        f"""\
private static String getEventTypeAsString(XMLEvent event) {{
{I}switch (event.getEventType()) {{
{II}case XMLStreamConstants.START_ELEMENT:
{III}return "Start-Element";
{II}case XMLStreamConstants.END_ELEMENT:
{III}return "End-Element";
{II}case XMLStreamConstants.PROCESSING_INSTRUCTION:
{III}return "Processing-Instruction";
{II}case XMLStreamConstants.CHARACTERS:
{III}return "Characters";
{II}case XMLStreamConstants.COMMENT:
{III}return "Comment";
{II}case XMLStreamConstants.SPACE:
{III}return "Space";
{II}case XMLStreamConstants.START_DOCUMENT:
{III}return "Start-Document";
{II}case XMLStreamConstants.END_DOCUMENT:
{III}return "End-Document";
{II}case XMLStreamConstants.ENTITY_REFERENCE:
{III}return "Entity-Reference";
{II}case XMLStreamConstants.ATTRIBUTE:
{III}return "Attribute";
{II}case XMLStreamConstants.NOTATION_DECLARATION:
{III}return "Notation-Declaration";
{II}default:
{III}return "Unknown-Type";
{I}}}
}}"""
    )


def _generate_try_element_name() -> Stripped:
    """Generate the function to strip the prefix and check the namespace."""
    return Stripped(
        f"""\
private static boolean invalidNameSpace(XMLEvent event) {{
{I}if (event.isStartElement()) {{
{II}return !AAS_NAME_SPACE.equals(event.asStartElement().getName().getNamespaceURI());
{I}}} else {{
{II}return !AAS_NAME_SPACE.equals(event.asEndElement().getName().getNamespaceURI());
{I}}}
}}

/**
 * Check the namespace and extract the element's name.
 */
private static Result<String> tryElementName(XMLEventReader reader) {{
{I}final XMLEvent currentEvent = currentEvent(reader);
{I}final boolean precondition = currentEvent.isStartElement() || currentEvent.isEndElement();
{I}if (!precondition) {{
{II}throw new IllegalStateException("Expected to be at a start or an end element "
{IIII}+ "but got: " + getEventTypeAsString(currentEvent));
{I}}}

{I}if (invalidNameSpace(currentEvent)) {{
{II}String namespace = currentEvent.isStartElement()
{IIII}? currentEvent.asStartElement().getName().getNamespaceURI()
{IIII}: currentEvent.asEndElement().getName().getNamespaceURI();
{II}final Reporting.Error error = new Reporting.Error(
{IIII}"Expected an element within a namespace " +
{IIII}AAS_NAME_SPACE + ", " + "but got: " + namespace);
{II}return Result.failure(error);
{I}}}
{I}return Result.success(currentEvent.isStartElement()
{III}? currentEvent.asStartElement().getName().getLocalPart()
{III}: currentEvent.asEndElement().getName().getLocalPart());
}}"""
    )


def _generate_verify_closing_tag_for_class() -> Stripped:
    return Stripped(
        f"""\
private static boolean isWrongClosingTag(
{I}Result<String> tryElementName,
{I}Result<String> tryEndElementName) {{
{I}return !tryElementName.getResult().equals(tryEndElementName.getResult());
}}

private static Result<XMLEvent> verifyClosingTagForClass(
{I}String className,
{I}XMLEventReader reader,
{I}Result<String> tryElementName) {{
{I}final XMLEvent currentEvent = currentEvent(reader);
{I}if (currentEvent.isEndDocument()) {{
{II}final Reporting.Error error = new Reporting.Error(
{IIII}"Expected an XML end element to conclude a property of class " + className
{IIIIII}+ " with the element name " + tryElementName.getResult() + ", "
{IIIIII}+ "but got the end-of-file.");
{II}return Result.failure(error);
{I}}}

{I}if (!currentEvent.isEndElement()) {{
{II}final Reporting.Error error = new Reporting.Error(
{IIII}"Expected an XML end element to conclude a property of class " + className
{IIIIII}+ " with the element name " + tryElementName.getResult() + ", "
{IIIIII}+ "but got the node of type " + getEventTypeAsString(currentEvent)
{IIIIII}+ " with the value " + currentEvent);
{II}return Result.failure(error);
{I}}}
{I}final Result<String> tryEndElementName = tryElementName(reader);
{I}if (tryEndElementName.isError()) {{
{II}return tryEndElementName.castTo(XMLEvent.class);
{I}}}
{I}if (isWrongClosingTag(tryElementName, tryEndElementName)) {{
{II}final Reporting.Error error = new Reporting.Error(
{IIII}"Expected an XML end element to conclude a property of class " + className
{IIIIII}+ " with the element name " + tryElementName.getResult() + ", "
{IIIIII}+ "but got the end element with the name " + tryEndElementName.getResult());
{II}return Result.failure(error);
{I}}}
{I}try {{
{II}return Result.success(reader.nextEvent());
{I}}} catch (XMLStreamException xmlStreamException) {{
{II}throw new Xmlization.DeserializeException("",
{III}"Failed in method verifyClosingTagForClass because of: " +
{III}xmlStreamException.getMessage());
{I}}}
}}"""
    )


def _generate_deserialize_enumeration_property(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> Stripped:
    """Generate the snippet to deserialize a property ``prop`` as an enum."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    assert isinstance(type_anno, intermediate.OurTypeAnnotation)

    our_type = type_anno.our_type
    assert isinstance(our_type, intermediate.Enumeration)

    target_var = java_naming.variable_name(Identifier(f"the_{prop.name}"))
    text_target_var = java_naming.variable_name(Identifier(f"text_{prop.name}"))
    optional_target_var = java_naming.variable_name(Identifier(f"optional_{prop.name}"))
    cls_name = java_naming.class_name(cls.name)
    prop_name = java_naming.property_name(prop.name)
    prop_type_name = java_naming.enum_name(our_type.name)
    from_str_name = java_naming.private_property_name(
        Identifier(f"{our_type.name}_from_string")
    )
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    return Stripped(
        f"""\
if (isEmptyProperty) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"The property {prop_name} of an instance of class {cls_name} " +
{II}"can not be de-serialized from a self-closing element " +
{II}"since it needs content");
{I}error.prependSegment(
{II}new Reporting.NameSegment(
{III}{xml_prop_name_literal}));
{I}return Result.failure(error);
}}

if (currentEvent(reader).isEndDocument()) {{
{I}final Reporting.Error error = new Reporting.Error(
{III}"Expected an XML content representing "
{IIIII}+ "the property {prop_name} of an instance of class {cls_name}, "
{IIIII}+ "but reached the end-of-file");
{I}return Result.failure(error);
}}

String {text_target_var};
try {{
{I}{text_target_var} = readContentAsString(reader);
}} catch (Exception e) {{
{I}final Reporting.Error error = new Reporting.Error(
{III}"The property {prop_name} of an instance of class {cls_name}"
{IIIII}+ " could not be de-serialized: " + e.getMessage());
{I}error.prependSegment(
{III}new Reporting.NameSegment(
{IIIII}"{prop_name}"));
{I}return Result.failure(error);
}}

final Optional<{prop_type_name}> {optional_target_var} =
{I}Stringification.{from_str_name}(
{II}{text_target_var});

if ({optional_target_var}.isPresent()) {{
{I}{target_var} = {optional_target_var}.get();
}} else {{
{I}final Reporting.Error error = new Reporting.Error(
{III}"The property {prop_name} of an instance of class {cls_name}" +
{IIIII}" could not be de-serialized from an unexpected enumeration literal: " +
{IIIII}{text_target_var});
{I}error.prependSegment(
{III}new Reporting.NameSegment(
{IIIII}"{prop_name}"));
{I}return Result.failure(error);
}}"""
    )


def _generate_deserialize_interface_property(
    prop: intermediate.Property,
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """Generate the snippet to deserialize a property ``prop`` as an interface."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    assert isinstance(type_anno, intermediate.OurTypeAnnotation)

    our_type = type_anno.our_type
    assert isinstance(
        our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
    )
    assert our_type.interface is not None

    prop_name = java_naming.property_name(prop.name)
    cls_name = java_naming.class_name(cls.name)

    interface_name = java_naming.interface_name(our_type.interface.name)

    target_var = java_naming.variable_name(Identifier(f"the_{prop.name}"))
    try_target_var = java_naming.variable_name(Identifier(f"try_{prop.name}"))
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    return Stripped(
        f"""\
if (isEmptyProperty) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"Expected an XML element within the element " + tryElementName.getResult() + " representing " +
{II}"the property {prop_name} of an instance of class {cls_name}, " +
{II}"but encountered a self-closing element.");
{I}return Result.failure(error);
}}

// We need to skip the whitespace here in order to be able to look ahead
// the discriminator element shortly.
skipWhitespaceAndComments(reader);

if (currentEvent(reader).isEndDocument()) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"Expected an XML element within the element " + tryElementName.getResult() + " representing " +
{II}"the property {prop_name} of an instance of class {cls_name}, " +
{II}"but reached the end-of-file");
{I}return Result.failure(error);
}}

// Try to look ahead the discriminator name;
// we need this name only for the error reporting below.
// {interface_name}FromElement will perform more sophisticated
// checks.
String discriminatorElementName = null;
if (currentEvent(reader).isStartElement()) {{
{I}Result<String> tryDiscriminatorElementName = tryElementName(reader);
{I}assert(!tryDiscriminatorElementName.isError());
{I}discriminatorElementName = tryDiscriminatorElementName.getResult();
}}

Result<? extends {interface_name}> {try_target_var} = try{interface_name}FromElement(reader);

if ({try_target_var}.isError()) {{
{I}if (discriminatorElementName != null) {{
{II}{try_target_var}.getError().
{III}prependSegment(
{IIII}new Reporting.NameSegment(
{IIIII}discriminatorElementName));
{I}}}

{I}{try_target_var}.getError()
{II}.prependSegment(
{III}new Reporting.NameSegment(
{IIII}{xml_prop_name_literal}));
{I}return {try_target_var}.castTo({cls_name}.class);
}}

{target_var} = {try_target_var}.getResult();"""
    )


def _generate_deserialize_cls_property(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> Stripped:
    """Generate the snippet to deserialize a property ``prop`` as a concrete class."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    assert isinstance(type_anno, intermediate.OurTypeAnnotation)

    our_type = type_anno.our_type
    assert isinstance(our_type, intermediate.ConcreteClass)

    target_cls_name = java_naming.class_name(our_type.name)

    target_var = java_naming.variable_name(Identifier(f"the_{prop.name}"))

    try_target_var = java_naming.variable_name(Identifier(f"try_{prop.name}"))

    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    cls_name = java_naming.class_name(cls.name)

    return Stripped(
        f"""\
Result<{target_cls_name}> {try_target_var} = try{target_cls_name}FromSequence(
{I}reader, isEmptyProperty);

if ({try_target_var}.isError()) {{
{I}{try_target_var}.getError()
{II}.prependSegment(
{III}new Reporting.NameSegment(
{IIII}{xml_prop_name_literal}));
{I}return {try_target_var}.castTo({cls_name}.class);
}}

{target_var} = {try_target_var}.getResult();"""
    )


def _generate_deserialize_list_property(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> Stripped:
    """Generate the code to de-serialize a property ``prop`` as a list."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # fmt: off
    assert (
        isinstance(type_anno, intermediate.ListTypeAnnotation)
        and isinstance(type_anno.items, intermediate.OurTypeAnnotation)
        and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass)
        )
    ), "See intermediate._translate._verify_only_simple_type_patterns"
    # fmt: on

    target_var = java_naming.variable_name(Identifier(f"the_{prop.name}"))

    item_our_type = type_anno.items.our_type

    if (
        isinstance(item_our_type, intermediate.AbstractClass)
        or len(item_our_type.concrete_descendants) > 0
    ):
        interface_name = java_naming.interface_name(type_anno.items.our_type.name)
        deserialize_method = f"{interface_name}FromElement"
    else:
        class_name = java_naming.class_name(type_anno.items.our_type.name)
        deserialize_method = f"{class_name}FromElement"

    item_type = java_common.generate_type(type_anno.items)

    cls_name = java_naming.class_name(cls.name)

    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    return Stripped(
        f"""\
{target_var} = new ArrayList<>();
if (!isEmptyProperty) {{
{I}skipWhitespaceAndComments(reader);
{I}int index = 0;
{I}if(!currentEvent(reader).isStartElement()){{
{II}final Reporting.Error error = new Reporting.Error(
{II}"Expected a start element opening an instance of {item_type}, but got an XML "
{III}+ getEventTypeAsString(currentEvent(reader)));
{II}error.prependSegment(new Reporting.IndexSegment(index));
{II}error.prependSegment(new Reporting.NameSegment({xml_prop_name_literal}));
{II}return Result.failure(error);
{I}}}
{I}while (currentEvent(reader).isStartElement()) {{

{II}Result<? extends {item_type}> itemResult = try{deserialize_method}(reader);

{II}if (itemResult.isError()) {{
{III}itemResult.getError()
{IIII}.prependSegment(
{IIIII}new Reporting.IndexSegment(index));
{III}itemResult.getError()
{IIII}.prependSegment(
{IIIII}new Reporting.NameSegment({xml_prop_name_literal}));
{III}return itemResult.castTo({cls_name}.class);
{II}}}

{II}{target_var}.add(itemResult.getResult());
{II}index++;
{II}skipWhitespaceAndComments(reader);
{I}}}
}}"""
    )


@require(lambda prop, cls: id(prop) in cls.property_id_set)
def _generate_deserialize_property(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the snippet to deserialize the property ``prop`` from the content."""
    blocks = []  # type: List[Stripped]

    type_anno = intermediate.beneath_optional(prop.type_annotation)

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        blocks.append(_generate_deserialize_primitive_property(prop=prop, cls=cls))
    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        our_type = type_anno.our_type
        if isinstance(our_type, intermediate.Enumeration):
            blocks.append(
                _generate_deserialize_enumeration_property(prop=prop, cls=cls)
            )
        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            # NOTE (empwilli, 2023-12-18):
            # The constrained primitives are only verified, but not represented as
            # separate classes in the XSD.
            blocks.append(_generate_deserialize_primitive_property(prop=prop, cls=cls))
        elif isinstance(
            our_type, (intermediate.ConcreteClass, intermediate.AbstractClass)
        ):
            if (
                isinstance(our_type, intermediate.AbstractClass)
                or len(our_type.concrete_descendants) > 0
            ):
                blocks.append(
                    _generate_deserialize_interface_property(prop=prop, cls=cls)
                )
            else:
                blocks.append(_generate_deserialize_cls_property(prop=prop, cls=cls))
        else:
            assert_never(our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        blocks.append(_generate_deserialize_list_property(prop=prop, cls=cls))

    else:
        assert_never(type_anno)

    return Stripped("\n\n".join(blocks)), None


def _generate_deserialize_impl_cls_from_sequence(
    cls: intermediate.ConcreteClass,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the function to de-serialize the ``cls`` from an XML sequence."""
    name = java_naming.class_name(identifier=cls.name)

    description = Stripped(
        f"""\
/**
 * Deserialize an instance of class {name} from a sequence of XML elements.
 *
 * <p>If {{@code isEmptySequence}} is set, we should try to deserialize
 * the instance from an empty sequence. That is, the parent element
 * was a self-closing element.
 */"""
    )

    # NOTE (empwilli, 2023-12-18):
    # Hard-wire for the case when no sequence is read
    if len(cls.constructor.arguments) == 0:
        return (
            Stripped(
                f"""\
{description}
private static Result<{name}> try{name}FromSequence(
{I}XMLEventReader reader,
{I}boolean isEmptySequence) {{
{I}return Result.success(new {name}());
}}"""
            ),
            None,
        )

    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    assert len(cls.constructor.arguments) > 0, "Otherwise expected hard-wiring above"
    init_target_var_stmts = []  # type: List[Stripped]
    for prop in cls.properties:
        type_anno = intermediate.beneath_optional(prop.type_annotation)

        assert isinstance(
            type_anno,
            (
                intermediate.PrimitiveTypeAnnotation,
                intermediate.OurTypeAnnotation,
                intermediate.ListTypeAnnotation,
            ),
        )

        target_type = java_common.generate_type(type_anno)
        target_var = java_naming.variable_name(Identifier(f"the_{prop.name}"))

        init_target_var_stmts.append(Stripped(f"{target_type} {target_var} = null;"))
    blocks.append(Stripped("\n".join(init_target_var_stmts)))

    # noinspection PyListCreation
    blocks_for_non_empty = []  # type: List[Stripped]

    blocks_for_non_empty.append(
        Stripped(
            f"""\
skipWhitespaceAndComments(reader);
if (currentEvent(reader).isEndDocument()) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"Expected an XML element representing " +
{II}"a property of an instance of class {name}, " +
{II}"but reached the end-of-file");
{I}return Result.failure(error);
}}"""
        )
    )

    case_blocks = []  # type: List[Stripped]
    for prop in cls.properties:
        case_body, error = _generate_deserialize_property(prop=prop, cls=cls)
        if error is not None:
            errors.append(error)
            continue

        assert case_body is not None

        xml_prop_name = naming.xml_property(prop.name)
        xml_prop_name_literal = java_common.string_literal(xml_prop_name)
        case_blocks.append(
            Stripped(
                f"""\
case {xml_prop_name_literal}:
{{
{I}{indent_but_first_line(case_body, I)}
{I}break;
}}"""
            )
        )

    if len(errors) > 0:
        return None, errors

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}final Reporting.Error error = new Reporting.Error(
{II}"We expected properties of the class {name}, " +
{II}"but got an unexpected element " +
{II}"with the name " + elementName);
{I}return Result.failure(error);"""
        )
    )

    switch_body = "\n".join(case_blocks)

    blocks_for_non_empty.append(
        Stripped(
            f"""\
while (true) {{
{I}skipWhitespaceAndComments(reader);

{I}if (currentEvent(reader).isEndElement() || currentEvent(reader).isEndDocument()) {{
{II}break;
{I}}}

{I}if (!currentEvent(reader).isStartElement()) {{
{II}final Reporting.Error error = new Reporting.Error(
{III}"Expected an XML start element representing " +
{III}"a property of an instance of class {name}, " +
{III}"but got the node of type " + getEventTypeAsString(currentEvent(reader)) +
{III}" with the value " + currentEvent(reader));
{II}return Result.failure(error);
{I}}}

{I}final Result<String> tryElementName = tryElementName(reader);
{I}if (tryElementName.isError()) {{
{II}return tryElementName.castTo({name}.class);
{I}}}

{I}final boolean isEmptyProperty = isEmptyElement(reader);
{I}final String elementName = tryElementName.getResult();

{I}switch (tryElementName.getResult()) {{
{II}{indent_but_first_line(switch_body, II)}
{I}}}

{I}skipWhitespaceAndComments(reader);


{I}final Result<XMLEvent> checkEndElement = verifyClosingTagForClass(
{II}"{name}",
{II}reader,
{II}tryElementName);
{I}if (checkEndElement.isError()) return checkEndElement.castTo({name}.class);

}}"""
        )
    )

    body_for_non_empty_sequence = "\n".join(blocks_for_non_empty)
    blocks.append(
        Stripped(
            f"""\
if (!isEmptySequence) {{
{I}{indent_but_first_line(body_for_non_empty_sequence, I)}
}}"""
        )
    )

    # region Check that the mandatory properties have been set

    for prop in cls.properties:
        prop_java = java_naming.property_name(prop.name)
        target_var = java_naming.variable_name(Identifier(f"the_{prop.name}"))

        if not isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            blocks.append(
                Stripped(
                    f"""\
if ({target_var} == null) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"The required property {prop_java} has not been given " +
{II}"in the XML representation of an instance of class {name}");
{I}return Result.failure(error);
}}"""
                )
            )

    # endregion

    # region Pass in properties as arguments to the constructor

    property_names = [prop.name for prop in cls.properties]
    constructor_argument_names = [arg.name for arg in cls.constructor.arguments]

    # fmt: off
    assert (
            set(prop.name for prop in cls.properties)
            == set(arg.name for arg in cls.constructor.arguments)
    ), (
        f"Expected the properties to coincide with constructor arguments, "
        f"but they do not for {cls.name!r}:"
        f"{property_names=}, {constructor_argument_names=}"
    )
    # fmt: on

    init_writer = io.StringIO()
    init_writer.write(f"return Result.success(new {name}(\n")

    for i, arg in enumerate(cls.constructor.arguments):
        prop = cls.properties_by_name[arg.name]

        # NOTE (empwilli, 2023-12-18):
        # The argument to the constructor may be optional while the property might
        # be required, since we can set the default value in the body of the
        # constructor. However, we can not have an optional property and a required
        # constructor argument as we then would not know how to create the instance.

        if not (
            intermediate.type_annotations_equal(
                arg.type_annotation, prop.type_annotation
            )
            or intermediate.type_annotations_equal(
                intermediate.beneath_optional(arg.type_annotation),
                prop.type_annotation,
            )
        ):
            errors.append(
                Error(
                    arg.parsed.node,
                    f"Expected type annotation for property {prop.name!r} "
                    f"and constructor argument {arg.name!r} "
                    f"of the class {cls.name!r} to have matching types, "
                    f"but they do not: "
                    f"property type is {prop.type_annotation} "
                    f"and argument type is {arg.type_annotation}. "
                    f"Hence we do not know how to generate the call "
                    f"to the constructor in the XML de-serialization.",
                )
            )
            continue

        arg_var = java_naming.variable_name(Identifier(f"the_{arg.name}"))

        init_writer.write(f"{I}{arg_var}")

        if i < len(cls.constructor.arguments) - 1:
            init_writer.write(",\n")
        else:
            init_writer.write("));")

    if len(errors) > 0:
        return None, errors

    # endregion

    blocks.append(Stripped(init_writer.getvalue()))

    writer = io.StringIO()
    writer.write(
        f"""\
{description}
private static Result<{name}> try{name}FromSequence(
{I}XMLEventReader reader,
{I}boolean isEmptySequence) {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_deserialize_impl_concrete_cls_from_element(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """Generate the function to de-serialize a concrete ``cls`` from an XML element."""
    name = java_naming.class_name(cls.name)
    xml_name = naming.xml_class_name(cls.name)
    xml_name_literal = java_common.string_literal(xml_name)

    body = Stripped(
        f"""\
skipWhitespaceAndComments(reader);

final XMLEvent currentEvent = currentEvent(reader);
if (currentEvent.getEventType() == XMLStreamConstants.END_DOCUMENT) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"Expected an XML element representing an instance of class {name}, " +
{II}"but reached the end-of-file");
{I}return Result.failure(error);
}}

if (currentEvent.getEventType() != XMLStreamConstants.START_ELEMENT) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"Expected an XML element representing an instance of class {name}, " +
{II}"but got a node of type " + getEventTypeAsString(currentEvent) +
{II}" with value " + currentEvent);
{I}return Result.failure(error);
}}

final Result<String> tryElementName = tryElementName(reader);
if (tryElementName.isError()) {{
{I}return tryElementName.castTo({name}.class);
}}

final String elementName = tryElementName.getResult();
if (!{xml_name_literal}.equals(tryElementName.getResult())) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"Expected an element representing an instance of class {name} " +
{II}"with element name {xml_name}, but got: " + elementName);
{I}return Result.failure(error);
}}

final boolean isEmptyElement = isEmptyElement(reader);

Result<{name}> result = try{name}FromSequence(
{I}reader,
{I}isEmptyElement);
if (result.isError()) return result.castTo({name}.class);


final Result<XMLEvent> checkEndElement = verifyClosingTagForClass(
{I}"{name}",
{I}reader,
{I}tryElementName);
if (checkEndElement.isError()) return checkEndElement.castTo({name}.class);


return result;"""
    )

    return Stripped(
        f"""\
/**
 * Deserialize an instance of class {name} from an XML element.
 */
private static Result<{name}> try{name}FromElement(
{I}XMLEventReader reader) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_deserialize_impl_interface_from_element(
    interface: intermediate.Interface,
) -> Stripped:
    """Generate the function to de-serialize an ``interface`` from an XML element."""
    name = java_naming.interface_name(interface.name)

    blocks = [
        Stripped(
            f"""\
skipWhitespaceAndComments(reader);

final XMLEvent currentEvent = currentEvent(reader);
if (currentEvent.getEventType() == XMLStreamConstants.END_DOCUMENT) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"Expected an XML element, but reached end-of-file");
{I}return Result.failure(error);
}}

if (currentEvent.getEventType() != XMLStreamConstants.START_ELEMENT) {{
{I}final Reporting.Error error = new Reporting.Error(
{II}"Expected an XML element representing an instance of class {name}, " +
{II}"but got a node of type " + getEventTypeAsString(currentEvent) +
{II}" with value " + currentEvent);
{I}return Result.failure(error);
}}"""
        )
    ]  # type: List[Stripped]

    case_stmts = []  # type: List[Stripped]
    for implementer in interface.implementers:
        implementer_xml_name_literal = java_common.string_literal(
            naming.xml_class_name(implementer.name)
        )

        implementer_name = java_naming.class_name(implementer.name)

        case_stmts.append(
            Stripped(
                f"""\
case {implementer_xml_name_literal}:
{I}return try{implementer_name}FromElement(reader);"""
            )
        )

    case_stmts.append(
        Stripped(
            f"""\
default:
{I}final Reporting.Error error = new Reporting.Error(
{II}"Unexpected element with the name " + getEventTypeAsString(currentEvent));
{I}return Result.failure(error);"""
        )
    )

    switch_writer = io.StringIO()
    switch_writer.write(
        f"""\
Result<String> tryElementName = tryElementName(
{I}reader);
if (tryElementName.isError()) {{
{I}return tryElementName.castTo({name}.class);
}}

final String elementName = tryElementName.getResult();
switch (elementName) {{
"""
    )
    for i, case_stmt in enumerate(case_stmts):
        if i > 0:
            switch_writer.write("\n")
        switch_writer.write(textwrap.indent(case_stmt, I))

    switch_writer.write("\n}")

    blocks.append(Stripped(switch_writer.getvalue()))

    writer = io.StringIO()
    writer.write(
        f"""\
/**
 * Deserialize an instance of {name} from an XML element.
 */
private static Result<? extends {name}> try{name}FromElement(
{I}XMLEventReader reader) {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_deserialize_impl(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the implementation for deserialization functions."""
    blocks = [
        _generate_current_event(),
        _generate_get_event_type_as_string(),
        _generate_is_empty_element(),
        _generate_verify_closing_tag_for_class(),
        _generate_skip_whitespace_and_comments(),
        _generate_skip_start_document(),
        _generate_try_element_name(),
        _generate_try_content_for_primitives(),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    # NOTE (empwilli, 2023-12-18):
    # Enumerations are going to be directly deserialized using
    # ``Stringification``.

    # NOTE (empwilli, 2023-12-18):
    # Constrained primitives are only verified, but do not represent a C# type.

    for cls in symbol_table.classes:
        if cls.is_implementation_specific:
            implementation_keys = [
                specific_implementations.ImplementationKey(
                    f"Xmlization/DeserializeImplementation/"
                    f"{cls.name}_from_element.java"
                ),
                specific_implementations.ImplementationKey(
                    f"Xmlization/DeserializeImplementation/"
                    f"{cls.name}_from_sequence.java"
                ),
            ]

            for implementation_key in implementation_keys:
                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            cls.parsed.node,
                            f"The xmlization snippet is missing "
                            f"for the implementation-specific "
                            f"class {cls.name}: {implementation_key}",
                        )
                    )
                    continue
                else:
                    blocks.append(spec_impls[implementation_key])
        else:
            if isinstance(cls, intermediate.ConcreteClass):
                (
                    block,
                    generation_errors,
                ) = _generate_deserialize_impl_cls_from_sequence(cls=cls)
                if generation_errors is not None:
                    errors.append(
                        Error(
                            cls.parsed.node,
                            f"Failed to generate the XML deserialization code "
                            f"for the class {cls.name}",
                            generation_errors,
                        )
                    )
                else:
                    assert block is not None
                    blocks.append(block)

            if cls.interface is not None:
                blocks.append(
                    _generate_deserialize_impl_interface_from_element(
                        interface=cls.interface
                    )
                )

            if isinstance(cls, intermediate.ConcreteClass):
                blocks.append(
                    _generate_deserialize_impl_concrete_cls_from_element(cls=cls)
                )

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()

    writer.write(
        """\
/**
 * Implement the deserialization of meta-model classes from XML.
 *
 * <p>The implementation propagates an {@link Reporting.Error} instead of
 * relying on exceptions. Under the assumption that incorrect data is much less
 * frequent than correct data, this makes the deserialization more
 * efficient.
 *
 * <p>However, we do not want to force the client to deal with
 * the {@link Reporting.Error} class as this is not intuitive.
 * Therefore we distinguish the implementation, realized in
 * {@link DeserializeImplementation}, and the facade given in
 * {@link Deserialize} class.
 */
private static class DeserializeImplementation
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_deserialize_from(name: Identifier) -> Stripped:
    """Generate the facade method for deserialization of the class or interface."""
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(name))
    writer = io.StringIO()

    writer.write(
        f"""\
/**
 * Deserialize an instance of {name} from {{@code reader}}.
 *
 * @param reader Initialized XML reader with reader.peek() set to the element
 */
"""
    )
    writer.write(
        f"""\
public static {name} deserialize{name}(
{I}XMLEventReader reader) {{

{I}DeserializeImplementation.skipStartDocument(reader);
{I}DeserializeImplementation.skipWhitespaceAndComments(reader);

{I}Result<? extends {name}> result =
{II}DeserializeImplementation.try{name}FromElement(
{III}reader);

{I}return result.onError(error -> {{
{II}error.prependSegment(new Reporting.NameSegment({xml_prop_name_literal}));
{II}throw new DeserializeException(
{III}Reporting.generateRelativeXPath(error.getPathSegments()),
{III}error.getCause());
{I}}});
}}"""
    )

    return Stripped(writer.getvalue())


def _generate_deserialize(symbol_table: intermediate.SymbolTable) -> Stripped:
    """Generate the public class ``Deserialize``."""
    blocks = []  # type: List[Stripped]

    # NOTE (empwilli, 2023-12-18):
    # We use stringification for de-serialization of enumerations.

    # NOTE (empwilli, 2023-12-18):
    # Constrained primitives are not handled as separate classes, but as
    # primitives, and only verified in the verification.

    for cls in symbol_table.classes:
        if cls.interface is not None:
            blocks.append(
                _generate_deserialize_from(
                    name=java_naming.interface_name(cls.interface.name)
                )
            )

        if isinstance(cls, intermediate.ConcreteClass):
            blocks.append(
                _generate_deserialize_from(name=java_naming.class_name(cls.name))
            )

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Deserialize instances of meta-model classes from XML.
 */
"""
    )

    first_cls = symbol_table.classes[0] if len(symbol_table.classes) > 0 else None

    if first_cls is not None:
        cls_name = None  # type: Optional[str]
        if isinstance(first_cls, intermediate.AbstractClass):
            cls_name = java_naming.interface_name(first_cls.name)
        elif isinstance(first_cls, intermediate.ConcreteClass):
            cls_name = java_naming.class_name(first_cls.name)
        else:
            assert_never(first_cls)

        an_instance_variable = java_naming.variable_name(Identifier("an_instance"))

        writer.write(
            f"""\
/** <pre>
 * Here is an example how to parse an instance of class {cls_name}:
 * {{@code
 * XMLEventReader reader = xmlFactory.createXMLEventReader(...some arguments...);
 * {cls_name} {an_instance_variable} = Deserialize.deserialize{cls_name}(
 * {I}reader);
 * }}
 * </pre>
 *
 * <pre>
 * If the elements live in a namespace, you have to supply it. For example:
 * {{@code
 * XMLEventReader reader = xmlFactory.createXMLEventReader(...some arguments...);
 * {cls_name} {an_instance_variable} = Deserialize.deserialize{cls_name}(
 * {I}reader,
 * {I}"http://www.example.com/5/12");
 * }}
 * </pre>
 */
"""
        )

    writer.write(
        """\
public static class Deserialize
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_serialize_primitive_property_as_content(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of the primitive-type ``prop`` as XML content."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    a_type = intermediate.try_primitive_type(type_anno)
    assert (
        a_type is not None
    ), f"Unexpected non-primitive type of the property {prop.name!r}: {type_anno}"

    prop_name = java_naming.property_name(prop.name)
    getter_name = java_naming.getter_name(prop.name)
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    write_value_block: Stripped

    if (
        a_type is intermediate.PrimitiveType.BOOL
        or a_type is intermediate.PrimitiveType.INT
        or a_type is intermediate.PrimitiveType.FLOAT
        or a_type is intermediate.PrimitiveType.STR
    ):
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            write_value_block = Stripped(
                f"""\
if (that.{getter_name}().isPresent()) {{
{I}writer.writeStartElement(
{II}{xml_prop_name_literal});
{I}if (topLevel) {{
{II}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{II}topLevel = false;
{I}}}

{I}writer.writeCharacters(
{II}that.{getter_name}().get().toString());

{I}writer.writeEndElement();
}}"""
            )
        else:
            write_value_block = Stripped(
                f"""\
writer.writeStartElement(
{I}{xml_prop_name_literal});
if (topLevel) {{
{I}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{I}topLevel = false;
}}
writer.writeCharacters(
{I}that.{getter_name}().toString());
writer.writeEndElement();"""
            )
    elif a_type is intermediate.PrimitiveType.BYTEARRAY:
        if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
            base64_prop_name = java_naming.property_name(
                Identifier(f"the_b64_{prop_name}")
            )
            write_value_block = Stripped(
                f"""\
if (that.{getter_name}().isPresent()) {{
{I}writer.writeStartElement({xml_prop_name_literal});
{I}if (topLevel) {{
{II}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{II}topLevel = false;
{I}}}
{I}String {base64_prop_name} = Base64.getEncoder().encodeToString(
{II}that.{getter_name}().get());
{I}writer.writeCharacters({base64_prop_name});
{I}writer.writeEndElement();
}}"""
            )
        else:
            base64_prop_name = java_naming.property_name(
                Identifier(f"the_b64_{prop_name}")
            )
            write_value_block = Stripped(
                f"""\
String {base64_prop_name} = Base64.getEncoder().encodeToString(
{I}that.{getter_name}());
writer.writeCharacters({base64_prop_name});"""
            )
    else:
        assert_never(a_type)

    assert write_value_block is not None

    write_value_block = Stripped(
        f"""\
try {{
{I}{indent_but_first_line(write_value_block, I)}
}} catch (Exception exception) {{
{I}throw new SerializeException("",exception.getMessage());
}}"""
    )

    return write_value_block


def _generate_serialize_enumeration_property_as_content(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of an enumeration ``prop`` as XML content."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)
    assert isinstance(type_anno, intermediate.OurTypeAnnotation) and isinstance(
        type_anno.our_type, intermediate.Enumeration
    ), "See intermediate._translate._verify_only_simple_type_patterns"

    enumeration = type_anno.our_type

    getter_name = java_naming.getter_name(prop.name)
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    enum_name = java_naming.enum_name(enumeration.name)

    text_var = java_naming.variable_name(Identifier(f"text_{prop.name}"))

    write_value_block: Stripped

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        write_value_block = Stripped(
            f"""\
if (that.{getter_name}().isPresent()) {{
{I}writer.writeStartElement(
{II}{xml_prop_name_literal});
{I}if (topLevel) {{
{II}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{II}topLevel = false;
{I}}}

{I}Optional<String> {text_var} = Stringification.toString(
{II}that.{getter_name}().get());

{I}if (!{text_var}.isPresent()) {{
{II}throw new IllegalArgumentException(
{III}"Invalid literal for the enumeration {enum_name}: " +
{III}that.{getter_name}().get().toString());
{I}}}

{I}writer.writeCharacters({text_var}.get());

{I}writer.writeEndElement();
}}"""
        )
    else:
        write_value_block = Stripped(
            f"""\
writer.writeStartElement(
{I}{xml_prop_name_literal});
if (topLevel) {{
{I}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{I}topLevel = false;
}}

Optional<String> {text_var} = Stringification.toString(
{I}that.{getter_name}());

if (!{text_var}.isPresent()) {{
{I}throw new IllegalArgumentException(
{II}"Invalid literal for the enumeration {enum_name}: " +
{II}that.{getter_name}().toString());
}}

writer.writeCharacters({text_var}.get());

writer.writeEndElement();"""
        )

    write_value_block = Stripped(
        f"""\
try {{
{I}{indent_but_first_line(write_value_block, I)}
}} catch (Exception exception) {{
{I}throw new SerializeException("",exception.getMessage());
}}"""
    )

    return write_value_block


def _generate_serialize_interface_property_as_content(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of an interface as XML content."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # fmt: off
    assert (
        isinstance(type_anno, intermediate.OurTypeAnnotation)
        and (
            isinstance(type_anno.our_type, intermediate.AbstractClass)
            or (
                isinstance(type_anno.our_type, intermediate.ConcreteClass)
                and len(type_anno.our_type.concrete_descendants) > 0
            )
        )
    ), "See intermediate._translate._verify_only_simple_type_patterns"
    # fmt: on

    getter_name = java_naming.getter_name(prop.name)
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    result: Stripped

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        result = Stripped(
            f"""\
if (that.{getter_name}().isPresent()) {{
{I}writer.writeStartElement(
{II}{xml_prop_name_literal});
{I}if (topLevel) {{
{II}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{II}topLevel = false;
{I}}}

{I}this.visit(
{II}that.{getter_name}().get(),
{II}writer);

{I}writer.writeEndElement();
}}"""
        )
    else:
        result = Stripped(
            f"""\
writer.writeStartElement(
{I}{xml_prop_name_literal});
if (topLevel) {{
{I}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{I}topLevel = false;
}}

this.visit(
{I}that.{getter_name}(),
{I}writer);

writer.writeEndElement();"""
        )

    result = Stripped(
        f"""\
try {{
{I}{indent_but_first_line(result, I)}
}} catch (XMLStreamException exception) {{
{I}throw new SerializeException("",exception.getMessage());
}}"""
    )

    return result


def _generate_serialize_concrete_class_property_as_sequence(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of the class ``prop`` as a sequence of properties."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)
    assert isinstance(type_anno, intermediate.OurTypeAnnotation)
    assert isinstance(type_anno.our_type, intermediate.ConcreteClass)

    cls_to_sequence = java_naming.method_name(
        Identifier(f"{type_anno.our_type.name}_to_sequence")
    )

    getter_name = java_naming.getter_name(prop.name)
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    result: Stripped

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        result = Stripped(
            f"""\
if (that.{getter_name}().isPresent()) {{
{I}writer.writeStartElement(
{II}{xml_prop_name_literal});
{I}if (topLevel) {{
{II}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{II}topLevel = false;
{I}}}

{I}this.{cls_to_sequence}(
{II}that.{getter_name}().get(),
{II}writer);

{I}writer.writeEndElement();
}}"""
        )
    else:
        result = Stripped(
            f"""\
writer.writeStartElement(
{I}{xml_prop_name_literal});
if (topLevel) {{
{I}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{I}topLevel = false;
}}

this.{cls_to_sequence}(
{I}that.{getter_name}(),
{I}writer);

writer.writeEndElement();"""
        )

    result = Stripped(
        f"""\
try {{
{I}{indent_but_first_line(result, I)}
}} catch (XMLStreamException exception) {{
{I}throw new SerializeException("",exception.getMessage());
}}"""
    )

    return result


def _generate_serialize_list_property_as_content(
    prop: intermediate.Property,
) -> Stripped:
    """Generate the serialization of a list ``prop`` as a sequence of elements."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    # fmt: off
    assert (
        isinstance(type_anno, intermediate.ListTypeAnnotation)
        and isinstance(type_anno.items, intermediate.OurTypeAnnotation)
        and isinstance(
            type_anno.items.our_type,
            (intermediate.AbstractClass, intermediate.ConcreteClass)
        )
    ), "See intermediate._translate._verify_only_simple_type_patterns"
    # fmt: on

    getter_name = java_naming.getter_name(prop.name)
    xml_prop_name_literal = java_common.string_literal(naming.xml_property(prop.name))

    result: Stripped

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        result = Stripped(
            f"""\
if (that.{getter_name}().isPresent()) {{
{I}writer.writeStartElement(
{I}{xml_prop_name_literal});
{I}if (topLevel) {{
{II}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{II}topLevel = false;
{I}}}
{I}for (IClass item : that.{getter_name}().get()) {{
{II}this.visit(
{III}item,
{III}writer);
{II}}}
{I}writer.writeEndElement();
}}"""
        )
    else:
        result = Stripped(
            f"""\
writer.writeStartElement(
{I}{xml_prop_name_literal});
if (topLevel) {{
{I}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{I}topLevel = false;
}}

for (IClass item : that.{getter_name}()) {{
{I}this.visit(
{II}item,
{II}writer);
}}

writer.writeEndElement();"""
        )

    result = Stripped(
        f"""\
try {{
{I}{indent_but_first_line(result, I)}
}} catch (XMLStreamException exception) {{
{I}throw new SerializeException("",exception.getMessage());
}}"""
    )

    return result


def _generate_serialize_property_as_content(prop: intermediate.Property) -> Stripped:
    """Generate the code to serialize the ``prop`` as content of an XML element."""
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    body = None  # type: Optional[Stripped]

    if isinstance(type_anno, intermediate.PrimitiveTypeAnnotation):
        body = _generate_serialize_primitive_property_as_content(prop=prop)
    elif isinstance(type_anno, intermediate.OurTypeAnnotation):
        our_type = type_anno.our_type

        if isinstance(our_type, intermediate.Enumeration):
            body = _generate_serialize_enumeration_property_as_content(prop=prop)

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            body = _generate_serialize_primitive_property_as_content(prop=prop)

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if (
                isinstance(our_type, intermediate.AbstractClass)
                or len(our_type.concrete_descendants) > 0
            ):
                body = _generate_serialize_interface_property_as_content(prop=prop)
            else:
                body = _generate_serialize_concrete_class_property_as_sequence(
                    prop=prop
                )

        else:
            assert_never(our_type)

    elif isinstance(type_anno, intermediate.ListTypeAnnotation):
        body = _generate_serialize_list_property_as_content(prop=prop)

    else:
        assert_never(type_anno)

    return body


def _generate_class_to_sequence(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the method to write ``cls`` as a sequence of properties as XML."""
    blocks = []  # type: List[Stripped]

    for prop in cls.properties:
        body = _generate_serialize_property_as_content(prop=prop)
        blocks.append(body)

    interface_name = java_naming.interface_name(cls.name)
    method_name = java_naming.method_name(Identifier(f"{cls.name}_to_sequence"))

    writer = io.StringIO()

    if len(cls.properties) == 0:
        blocks.append(Stripped("// Intentionally empty."))

    writer.write(
        f"""\
private void {method_name}(
{I}{interface_name} that,
{I}XMLStreamWriter writer) {{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


def _generate_visit_for_class(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the method to write the ``cls`` as an XML element."""
    interface_name = java_naming.interface_name(cls.name)
    visit_name = java_naming.method_name(Identifier(f"visit_{cls.name}"))

    cls_to_sequence_name = java_naming.method_name(
        Identifier(f"{cls.name}_to_sequence")
    )

    xml_cls_name_literal = java_common.string_literal(naming.xml_class_name(cls.name))

    writer = io.StringIO()

    writer.write(
        f"""\
@Override
public void {visit_name}(
{I}{interface_name} that,
{I}XMLStreamWriter writer) {{
{I}try {{
{II}writer.writeStartElement(
{III}{xml_cls_name_literal});
{II}if (topLevel) {{
{III}writer.writeNamespace("xmlns", AAS_NAME_SPACE);
{III}topLevel = false;
{II}}}
{II}this.{cls_to_sequence_name}(
{III}that,
{III}writer);
{II}writer.writeEndElement();
{I}}} catch (XMLStreamException exception) {{
{II}throw new SerializeException("", exception.getMessage());
{I}}}
}}"""
    )

    return Stripped(writer.getvalue())


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_visitor(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate a visitor which serializes instances of the meta-model to XML."""
    errors = []  # type: List[Error]

    blocks = []  # type: List[Stripped]

    # The abstract classes are directly dispatched by the transformer,
    # so we do not need to handle them separately.

    for cls in symbol_table.concrete_classes:
        if cls.is_implementation_specific:
            implementation_keys = [
                specific_implementations.ImplementationKey(
                    f"Xmlization/VisitorWithWriter/visit_{cls.name}.java"
                ),
                specific_implementations.ImplementationKey(
                    f"Xmlization/VisitorWithWriter/{cls.name}_to_sequence.java"
                ),
            ]

            for implementation_key in implementation_keys:
                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            cls.parsed.node,
                            f"The xmlization snippet is missing "
                            f"for the implementation-specific "
                            f"class {cls.name}: {implementation_key}",
                        )
                    )
                    continue

                blocks.append(spec_impls[implementation_key])
        else:
            blocks.append(_generate_class_to_sequence(cls=cls))

            blocks.append(_generate_visit_for_class(cls=cls))

    if len(errors) > 0:
        return None, errors

    writer = io.StringIO()
    writer.write(
        f"""\
/**
 * Serialize recursively the instances as XML elements.
 */
static class VisitorWithWriter
{I}extends AbstractVisitorWithContext<XMLStreamWriter> {{

{I}private boolean topLevel = true;

"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue()), None


def _generate_serialize(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """Generate the static serializer."""
    blocks = [
        Stripped(
            f"""\
/**
 * Serialize an instance of the meta-model to XML.
 */
public static void to(
{I}IClass that,
{I}XMLStreamWriter writer) throws SerializeException {{
{I}VisitorWithWriter visitor = new VisitorWithWriter();
{I}visitor.visit(
{II}that, writer);
}}"""
        ),
    ]  # type: List[Stripped]

    writer = io.StringIO()
    writer.write(
        """\
/**
 * Serialize instances of meta-model classes to XML.
 */
"""
    )

    first_cls = (
        symbol_table.classes[0] if len(symbol_table.classes) > 0 else None
    )  # type: Optional[intermediate.ClassUnion]

    if first_cls is not None:
        cls_name = None  # type: Optional[str]
        if isinstance(first_cls, intermediate.AbstractClass):
            cls_name = java_naming.interface_name(first_cls.name)
        elif isinstance(first_cls, intermediate.ConcreteClass):
            cls_name = java_naming.class_name(first_cls.name)
        else:
            assert_never(first_cls)

        an_instance_variable = java_naming.variable_name(Identifier("an_instance"))

        writer.write(
            f"""\
/**
 * <pre>
 * Here is an example how to serialize an instance of {cls_name}:
 * {{@code
 * IClass {an_instance_variable} = new {cls_name}(
 *     ... some constructor arguments ...
 * );
 * XMLStreamWriter writer = xmlWriterFactory.createXMLStreamWriter(...some arguments...);
 * Serialize.to(
 * {I}{an_instance_variable},
 * {I}writer);
 * }}
 * </pre>
 */
"""
        )

    writer.write(
        """\
public static class Serialize
{
"""
    )

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n}")

    return Stripped(writer.getvalue())


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable,
    package: java_common.PackageIdentifier,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the Java code for the general serialization.

    The ``package`` defines the root Java package.
    """
    errors = []  # type: List[Error]

    imports = [
        Stripped("import javax.xml.stream.events.XMLEvent;"),
        Stripped("import javax.xml.stream.XMLEventReader;"),
        Stripped("import javax.xml.stream.XMLStreamConstants;"),
        Stripped("import javax.xml.stream.XMLStreamException;"),
        Stripped("import javax.xml.stream.XMLStreamWriter;"),
        Stripped("import java.util.ArrayList;"),
        Stripped("import java.util.Base64;"),
        Stripped("import java.util.function.Function;"),
        Stripped("import java.util.List;"),
        Stripped("import java.util.Optional;"),
        Stripped(f"import {package}.reporting.Reporting;"),
        Stripped(f"import {package}.stringification.Stringification;"),
        Stripped(f"import {package}.types.enums.*;"),
        Stripped(f"import {package}.types.impl.*;"),
        Stripped(f"import {package}.types.model.*;"),
        Stripped(f"import {package}.visitation.*;"),
    ]  # type: List[Stripped]

    # region Deserialization helpers

    xml_result_class = _generate_result()

    xml_namespace_literal = java_common.string_literal(
        symbol_table.meta_model.xml_namespace
    )

    # endregion

    # region Deserialization Implementation

    deserialize_impl_block, deserialize_impl_errors = _generate_deserialize_impl(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if deserialize_impl_errors is not None:
        errors.extend(deserialize_impl_errors)

    assert deserialize_impl_block is not None

    # endregion

    # region Deserialization

    deserialize_block = _generate_deserialize(symbol_table=symbol_table)

    # endregion

    # region Visitor

    visitor_block, visitor_errors = _generate_visitor(
        symbol_table=symbol_table, spec_impls=spec_impls
    )
    if visitor_errors is not None:
        errors.extend(visitor_errors)

    assert visitor_block is not None

    # endregion

    # region Serialization

    serialization_block = _generate_serialize(symbol_table=symbol_table)

    # endregion

    xmlization_blocks = [
        Stripped(
            f"""\
/**
 * Provide de/serialization of meta-model classes to/from XML.
 */
public class Xmlization {{
{I}/**
{I} * Represent a critical error during the deserialization.
{I} */
{I}@SuppressWarnings("serial")
{I}public static class DeserializeException extends RuntimeException {{
{II}private final String path;
{II}private final String reason;

{II}public DeserializeException(String path, String reason) {{
{III}super(reason + " at: " + ("".equals(path) ? "the beginning" : path));
{III}this.path = path;
{III}this.reason = reason;
{II}}}

{II}public Optional<String> getPath() {{
{III}return Optional.ofNullable(path);
{II}}}

{II}public Optional<String> getReason() {{
{III}return Optional.ofNullable(reason);
{II}}}
{I}}}

{I}/**
{I} * Represent a critical error during the serialization.
{I} */
{I}@SuppressWarnings("serial")
{I}public static class SerializeException extends RuntimeException {{
{II}private final String path;
{II}private final String reason;

{II}public SerializeException(String path, String reason) {{
{III}super(reason + " at: " + ("".equals(path) ? "the beginning" : path));
{III}this.path = path;
{III}this.reason = reason;
{II}}}

{II}public Optional<String> getPath() {{
{III}return Optional.ofNullable(path);
{II}}}

{II}public Optional<String> getReason() {{
{III}return Optional.ofNullable(reason);
{II}}}
{I}}}

{I}/**
{I} * The XML namespace of the meta-model
{I} */
{I}public static final String AAS_NAME_SPACE =
{II}{xml_namespace_literal};

{I}{indent_but_first_line(xml_result_class, I)}

{I}{indent_but_first_line(deserialize_impl_block, I)}

{I}{indent_but_first_line(deserialize_block, I)}

{I}{indent_but_first_line(visitor_block, I)}

{I}{indent_but_first_line(serialization_block, I)}
}}"""
        ),
    ]  # type: List[Stripped]

    if len(errors) > 0:
        return None, errors

    blocks = [
        java_common.WARNING,
        Stripped(f"package {package}.xmlization;"),
        Stripped("\n".join(imports)),
        Stripped("\n\n".join(xmlization_blocks)),
        java_common.WARNING,
    ]  # type: List[Stripped]

    code = "\n\n".join(blocks)

    return f"{code}\n", None


# endregion



"""Generate the Golang code for reporting errors by including the code directly."""

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.golang import (
    common as golang_common,
)
from aas_core_codegen.golang.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate() -> str:
    """Generate the Golang code for reporting errors."""
    blocks = [
        Stripped(
            """\
// Package reporting provides structures and functions for reporting of errors.
package reporting"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"fmt"
{I}"strings"
{I}"strconv"
)"""
        ),
        Stripped(
            f"""\
type NameSegment struct{{
{I}Name string
}}"""
        ),
        Stripped(
            f"""\
type IndexSegment struct{{
{I}Index int
}}"""
        ),
        Stripped(
            f"""\
type Path struct{{
{I}// A segments is expected to be either a name segment or an index segment.
{I}segments []interface{{}}
{I}start int
}}"""
        ),
        Stripped(
            f"""\
// Prepend the segment to the path.
//
// Grow segments exponentially if there is no place.
func (p *Path) prepend(segment interface{{}}) {{
{I}// See: https://en.wikipedia.org/wiki/Amortized_analysis#Dynamic_array
{I}if len(p.segments) == 0 {{
{II}p.segments = make([]interface{{}}, 1)
{II}p.segments[0] = segment
{II}p.start = 0
{I}}} else if p.start > 0 {{
{II}p.start--
{II}p.segments[p.start] = segment
{I}}} else {{
{II}s := make([]interface{{}}, len(p.segments) * 2)
{II}copy(s[len(p.segments):], p.segments)
{II}p.start = len(p.segments) - 1
{II}p.segments = s
{II}p.segments[p.start] = segment
{I}}}
}}"""
        ),
        Stripped(
            f"""\
// Prepend the name segment to the path.
func (p *Path) PrependName(segment *NameSegment) {{
{I}p.prepend(segment)
}}"""
        ),
        Stripped(
            f"""\
// Prepend the index segment to the path.
func (p *Path) PrependIndex(segment *IndexSegment) {{
{I}p.prepend(segment)
}}"""
        ),
        Stripped(
            f"""\
// Apply the `callback` on each segment.
//
// The segment object is either a [NameSegment] or a [IndexSegment].
func (p *Path) OverSegments(callback func(interface{{}})) {{
{I}start := p.start
{I}for i := start; i < len(p.segments); i++ {{
{II}callback(p.segments[i])
{I}}}
}}"""
        ),
        Stripped(
            f"""\
// Translate the path to a JSON path.
//
// The name segments are expected to denote the names of the properties
// in JSON property names, not Golang property names.
func ToJSONPath(p *Path) string {{
{I}var b strings.Builder
{I}i := 0
{I}p.OverSegments(func(s interface{{}}) {{
{II}switch v := s.(type) {{
{II}case *NameSegment:
{III}if i == 0 {{
{III}{I}b.WriteString(v.Name)
{III}}} else {{
{III}{I}b.WriteString(".")
{III}{I}b.WriteString(v.Name)
{III}}}
{II}case *IndexSegment:
{III}b.WriteString("[")
{III}b.WriteString(strconv.Itoa(v.Index))
{III}b.WriteString("]")
{II}default:
{III}panic(
{III}{I}fmt.Sprintf(
{III}{II}"Unexpected segment of type %T: %v",
{III}{II}s, s,
{III}{I}),
{III})
{II}}}

{II}i++
{I}}})
{I}return b.String()
}}"""
        ),
        Stripped(
            f"""\
// Translate the path to a Golang access path.
//
// The name segments are expected to denote the names of the properties
// in Golang, not JSON property names.
func ToGolangPath(p *Path) string {{
{I}// NOTE(mristin):
{I}// We re-use JSON path formatting as implementation, but introduce
{I}// a separate function to signal to the reader in which form
{I}// the name segments are expected (Golang property names instead
{I}// of JSON property names).
{I}return ToJSONPath(p)
}}"""
        ),
        Stripped(
            f"""\
var replacerForXPath = strings.NewReplacer(
{I}"/", "&#47;",
{I}"<", "&lt;",
{I}">", "&gt;",
{I}"\\"", "&quot;",
{I}"'", "&apos;",
)"""
        ),
        Stripped(
            f"""\
// Escape special characters for XPath.
func escapeForXPath(
{I}text string,
) string {{
{I}return replacerForXPath.Replace(text)
}}"""
        ),
        Stripped(
            f"""\
// Generate a relative XPath based on the path segments.
//
// Leave out the leading slash (`/`). This is helpful if we
// want to embed the error report in a larger document with a prefix
// *etc.*
func ToRelativeXPath(
{I}p *Path,
) string {{
{I}i := 0
{I}var b strings.Builder
{I}p.OverSegments(func(s interface{{}}) {{
{II}if i > 0 {{
{III}b.WriteString("/")
{II}}}

{II}switch v := s.(type) {{
{III}case *NameSegment:
{IIII}b.WriteString(escapeForXPath(v.Name))
{III}case *IndexSegment:
{IIII}b.WriteString(fmt.Sprintf("*[%d]", v.Index))
{IIIII}{I}
{III}default:
{IIII}panic(fmt.Sprintf("Unexpected segment of type %T: %v", s, s))
{II}}}
{II}
{II}i++
{I}}})
{I}return b.String()
}}"""
        ),
        golang_common.WARNING,
    ]

    return "\n\n".join(blocks) + "\n"



"""Generate the shared code used across the xmlization unit tests."""

import io

from icontract import ensure

from aas_core_codegen.common import Stripped
from aas_core_codegen.python import (
    common as python_common,
)
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)


@ensure(
    lambda result: result.endswith("\n"),
    "Trailing newline mandatory for valid end-of-files",
)
def generate(qualified_module_name: python_common.QualifiedModuleName) -> str:
    """
    Generate the shared code used across the xmlization unit tests.

    The ``qualified_module_name`` indicates the fully-qualified name of the base module.
    """
    blocks = [
        Stripped(
            '"""Provide common functionality for XML de/serialization in tests."""',
        ),
        python_common.WARNING,
        Stripped(
            """\
import io
import pathlib
import sys
from typing import Iterator, Optional
import xml.etree.ElementTree as ET"""
        ),
        Stripped(
            """\
if sys.version_info >= (3, 8):
    from typing import Final
else:
    from typing_extensions import Final"""
        ),
        Stripped(
            f"""\
import {qualified_module_name}.types as aas_types
import {qualified_module_name}.xmlization as aas_xmlization"""
        ),
        Stripped(
            f'''\
class Difference:
{I}"""Represent a single difference between two XML documents."""

{I}#: Human-readable description of the difference
{I}message: Final[str]

{I}#: Path to the expected XML element which is different from
{I}#: the obtained XML element
{I}path: Final[aas_xmlization.Path]

{I}def __init__(self, message: str) -> None:
{II}"""Initialize with the given message and empty path."""
{II}self.message = message
{II}self.path = aas_xmlization.Path()

{I}def __str__(self) -> str:
{II}return f"#{{self.path}}: {{self.message}}"'''
        ),
        Stripped(
            f'''\
def remove_redundant_whitespace(element: ET.Element) -> None:
{I}"""
{I}Remove the whitespace which will be ignored in the XML parsing.

{I}:param element: to remove the whitespace from
{I}"""
{I}if len(element) > 0:
{II}if element.text is not None and len(element.text.strip()) == 0:
{III}element.text = None

{II}for child in element:
{III}if child.tail is not None and len(child.tail.strip()) == 0:
{IIII}child.tail = None

{III}remove_redundant_whitespace(child)'''
        ),
        Stripped(
            f'''\
def check_equal(
{I}expected: ET.Element,
{I}got: ET.Element,
) -> Iterator[Difference]:
{I}"""
{I}Compare recursively two XML elements for equality.

{I}Make sure you called :py:function:`.remove_redundant_whitespace`
{I}so that your comparison does not trip over trivial whitespace.

{I}:param expected: expected XML element
{I}:param got: obtained XML element
{I}:yield: differences
{I}"""
{I}# NOTE (mristin):
{I}# We need to ignore the white-space before the children as it is ignored
{I}# by the XML parser.

{I}stop_recursion = False
{I}if expected.text is not None and got.text is None:
{II}stop_recursion = True
{II}yield Difference(f"Expected text {{expected.text!r}}, but got none")

{I}if expected.text is None and got.text is not None:
{II}stop_recursion = True
{II}yield Difference(f"Expected no text, but got {{got.text!r}}")

{I}if expected.tail is not None and got.tail is None:
{II}stop_recursion = True
{II}yield Difference(f"Expected tail {{expected.tail!r}}, but got none")

{I}if expected.tail is None and got.tail is not None:
{II}stop_recursion = True
{II}yield Difference(f"Expected no tail, but got {{got.tail!r}}")

{I}if expected.text is not None and got.text is not None and expected.text != got.text:
{II}stop_recursion = True
{II}yield Difference(f"Expected text {{expected.text!r}}, but got {{got.text!r}}")

{I}if expected.tail is not None and got.tail is not None and expected.tail != got.tail:
{II}stop_recursion = True
{II}yield Difference(f"Expected tail {{expected.tail!r}}, but got {{got.tail!r}}")

{I}if expected.tag != got.tag:
{II}stop_recursion = True
{II}yield Difference(f"Expected tail {{expected.tag!r}}, but got {{got.tag!r}}")

{I}expected_children = [  # pylint: disable=unnecessary-comprehension
{II}child for child in expected
{I}]

{I}got_children = [child for child in got]  # pylint: disable=unnecessary-comprehension

{I}if len(expected_children) != len(got_children):
{II}stop_recursion = True
{II}yield Difference(
{III}f"Expected {{len(expected_children)}} child element(s), "
{III}f"but got {{len(got_children)}} child element(s)"
{II})

{I}children_tag_unique = len(set(child.tag for child in expected_children)) == len(
{II}expected_children
{I})

{I}if stop_recursion:
{II}return

{I}for i, (expected_child, got_child) in enumerate(
{II}zip(expected_children, got_children)
{I}):
{II}for difference in check_equal(expected_child, got_child):
{III}if children_tag_unique:
{IIII}difference.path._prepend(aas_xmlization.ElementSegment(expected_child))
{III}else:
{IIII}difference.path._prepend(aas_xmlization.IndexSegment(expected_child, i))

{III}yield difference'''
        ),
        Stripped(
            f'''\
def assert_elements_equal(
{I}expected: ET.Element, got: ET.Element, message_if_not_equal: Optional[str] = None
) -> None:
{I}"""
{I}Assert that the two elements are equal.

{I}Make sure you called :py:function:`.remove_redundant_whitespace`
{I}so that your comparison does not trip over trivial whitespace.

{I}:param expected: what you expected
{I}:param got: what you got
{I}:param message_if_not_equal: description or identifier to help debugging
{I}:raise: :py:class:`AssertionError` if the two elements are not equal
{I}"""
{I}findings_text = "\\n".join(
{II}str(difference) for difference in check_equal(expected=expected, got=got)
{I})

{I}if len(findings_text) != 0:
{II}writer = io.StringIO()
{II}writer.write(
{III}f"Expected two elements to be equal, but they are not. "
{III}f"Differences related to the expected element:\\n"
{III}f"{{findings_text}}"
{II})

{II}if message_if_not_equal is not None:
{III}writer.write("\\n\\n")
{III}writer.write(message_if_not_equal)

{II}raise AssertionError(writer.getvalue())'''
        ),
        Stripped(
            f'''\
def must_load(
{I}path: pathlib.Path,
) -> aas_types.Class:
{I}"""Load an instance from ``path``."""
{I}try:
{II}instance = aas_xmlization.from_file(path)
{I}except Exception as exception:
{II}raise RuntimeError(f"Failed to read from {{path}}") from exception

{I}return instance'''
        ),
        python_common.WARNING,
    ]

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue()



"""Generate Golang code of common functions by including the code directly."""

import io

from icontract import ensure

from aas_core_codegen.common import (
    Stripped,
)
from aas_core_codegen.golang import (
    common as golang_common,
)
from aas_core_codegen.golang.common import INDENT as I, INDENT2 as II, INDENT3 as III


# fmt: off
@ensure(
    lambda result:
    result.endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate() -> str:
    """Generate the Golang code for common functions."""
    blocks = [
        Stripped(
            """\
// Package common provides common functions shared among the other packages.
package common"""
        ),
        golang_common.WARNING,
        Stripped(
            f"""\
import (
{I}"strings"
)"""
        ),
        Stripped(
            f"""\
func Concat(
{I}parts ...string,
) string {{
{I}b := new(strings.Builder)
{I}for _, part := range parts {{
{II}b.WriteString(part)
{I}}}
{I}return b.String()
}}"""
        ),
        Stripped(
            f"""\
// Check if the map contains the given key.
func MapContains[K comparable, V any](m map[K]V, k K) bool {{
{I}_, ok := m[k]
{I}return ok
}}"""
        ),
        Stripped(
            f"""\
// Check if any of the elements satisfy the condition.
func Some[V any](condition func(V) bool, l []V) bool {{
{I}ok := false
{I}for _, v := range l {{
{II}ok = ok || condition(v)
{I}}}
{I}return ok
}}"""
        ),
        Stripped(
            f"""\
// Check if all the elements satisfy the condition.
func All[V any](condition func(V) bool, l []V) bool {{
{I}for _, v := range l {{
{II}if !condition(v) {{
{III}return false
{II}}}
{I}}}
{I}return true
}}"""
        ),
        Stripped(
            f"""\
// Check if some of the elements in the given range satisfy the condition.
func SomeRange(condition func(int) bool, start int, end int) bool {{
{I}for i := start; i < end; i++ {{
{II}if condition(i) {{
{III}return true
{II}}}
{I}}}
{I}return false
}}"""
        ),
        Stripped(
            f"""\
// Check if all the elements in the given range satisfy the condition.
func AllRange(condition func(int) bool, start int, end int) bool {{
{I}for i := start; i < end; i++ {{
{II}if !condition(i) {{
{III}return false
{II}}}
{I}}}
{I}return true
}}"""
        ),
        Stripped(
            f"""\
// Create a new instance of the `value` and return the pointer to it.
func NewBool(value bool) *bool {{
{I}return &value
}}"""
        ),
        Stripped(
            f"""\
// Create a new instance of the `value` and return the pointer to it.
func NewInt(value int) *int {{
{I}return &value
}}"""
        ),
        Stripped(
            f"""\
// Create a new instance of the `value` and return the pointer to it.
func NewInt64(value int64) *int64 {{
{I}return &value
}}"""
        ),
        Stripped(
            f"""\
// Create a new instance of the `value` and return the pointer to it.
func NewFloat64(value float64) *float64 {{
{I}return &value
}}"""
        ),
        Stripped(
            f"""\
// Create a new instance of the `value` and return the pointer to it.
func NewString(value string) *string {{
{I}return &value
}}"""
        ),
        golang_common.WARNING,
    ]

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue()



"""Provide common functions shared among different Go code generation modules."""
import io
import math
from typing import List, Tuple, Optional

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped, assert_never, Identifier
from aas_core_codegen.golang import (
    naming as golang_naming,
    pointering as golang_pointering,
)


@ensure(lambda result: result.startswith('"'))
@ensure(lambda result: result.endswith('"'))
def string_literal(text: str) -> Stripped:
    """Generate a Go string literal from the ``text``."""
    escaped = []  # type: List[str]

    for character in text:
        code_point = ord(character)

        if character == "\a":
            escaped.append("\\a")
        elif character == "\b":
            escaped.append("\\b")
        elif character == "\f":
            escaped.append("\\f")
        elif character == "\n":
            escaped.append("\\n")
        elif character == "\r":
            escaped.append("\\r")
        elif character == "\t":
            escaped.append("\\t")
        elif character == "\v":
            escaped.append("\\v")
        elif character == '"':
            escaped.append('\\"')
        elif character == "\\":
            escaped.append("\\\\")
        elif code_point < 32:
            # Non-printable ASCII characters
            escaped.append(f"\\x{ord(character):x}")
        elif 255 < code_point < 65536:
            # Above ASCII
            escaped.append(f"\\u{ord(character):04x}")
        elif code_point >= 65536:
            # Above Unicode Binary Multilingual Pane
            escaped.append(f"\\U{ord(character):08x}")
        else:
            escaped.append(character)

    return Stripped('"{}"'.format("".join(escaped)))


def needs_escaping(text: str) -> bool:
    """Check whether the ``text`` contains a character that needs escaping."""
    for character in text:
        if character == "\a":
            return True
        elif character == "\b":
            return True
        elif character == "\f":
            return True
        elif character == "\n":
            return True
        elif character == "\r":
            return True
        elif character == "\t":
            return True
        elif character == "\v":
            return True
        elif character == '"':
            return True
        elif character == "\\":
            return True
        else:
            pass

    return False


def boolean_literal(value: bool) -> Stripped:
    """Generate the boolean literal corresponding to the ``value``."""
    return Stripped("true") if value else Stripped("false")


def float_literal(value: float) -> Stripped:
    """Generate the float literal.

    We assume that the precision of the literal is not critical and rely on
    Python's ``str(.)`` function. However, if you want to specify the exact
    number, you have to format the number yourself, probably using G17 representation.
    """
    if math.isnan(value):
        return Stripped("math.NaN")
    if value == math.inf:
        return Stripped("math.Inf(1)")
    elif value == -math.inf:
        return Stripped("math.Inf(-1)")
    else:
        return Stripped(str(value))


# NOTE (mristin, 2023-01-13):
# See: https://stackoverflow.com/questions/19094704/indentation-in-go-tabs-or-spaces
INDENT = "\t"


def bytes_literal(value: bytes) -> Tuple[Stripped, bool]:
    """
    Generate an expression representing the ``value``.

    If there are more than 8 bytes, a multi-line expression is returned.

    :param value: to be represented
    :return: (Golang expression, is multi-line)
    """
    if len(value) == 0:
        return Stripped("[...]byte{}"), False

    writer = io.StringIO()

    if len(value) <= 8:
        items_joined = ", ".join(f"0x{byte:02x}" for byte in value)
        return Stripped(f"[...]byte{{{items_joined}}}"), False
    else:
        writer.write(
            """\
[...]byte {"""
        )

        for start in range(0, len(value), 8):
            if start == 0:
                writer.write(f"\n{INDENT}")
            else:
                writer.write(f",\n{INDENT}")

            end = min(start + 8, len(value))

            assert start < end

            for i, byte in enumerate(value[start:end]):
                if i > 0:
                    writer.write(", ")

                writer.write(f"0x{byte:02x}")

        writer.write("\n}")

        return Stripped(writer.getvalue()), True


PRIMITIVE_TYPE_MAP = {
    intermediate.PrimitiveType.BOOL: Stripped("bool"),
    intermediate.PrimitiveType.INT: Stripped("int64"),
    intermediate.PrimitiveType.FLOAT: Stripped("float64"),
    intermediate.PrimitiveType.STR: Stripped("string"),
    intermediate.PrimitiveType.BYTEARRAY: Stripped("[]byte"),
}


def _assert_all_primitive_types_are_mapped() -> None:
    """Assert that we have explicitly mapped all the primitive types to Go."""
    all_primitive_literals = set(literal.value for literal in PRIMITIVE_TYPE_MAP)

    mapped_primitive_literals = set(
        literal.value for literal in intermediate.PrimitiveType
    )

    all_diff = all_primitive_literals.difference(mapped_primitive_literals)
    mapped_diff = mapped_primitive_literals.difference(all_primitive_literals)

    messages = []  # type: List[str]
    if len(mapped_diff) > 0:
        messages.append(
            f"More primitive maps are mapped than there were defined "
            f"in the ``intermediate._types``: {sorted(mapped_diff)}"
        )

    if len(all_diff) > 0:
        messages.append(
            f"One or more primitive types in the ``intermediate._types`` were not "
            f"mapped in PRIMITIVE_TYPE_MAP: {sorted(all_diff)}"
        )

    if len(messages) > 0:
        raise AssertionError("\n\n".join(messages))


_assert_all_primitive_types_are_mapped()

TYPES_PACKAGE = Identifier("aastypes")
CONSTANTS_PACKAGE = Identifier("aasconstants")
VERIFICATION_PACKAGE = Identifier("aasverification")


def generate_type(
    type_annotation: intermediate.TypeAnnotationUnion,
    types_package: Optional[Identifier] = None,
) -> Stripped:
    """
    Generate the Go type for the given type annotation.

    If ``types_package`` is specified, it is prepended to all our types.
    """
    if isinstance(type_annotation, intermediate.PrimitiveTypeAnnotation):
        return PRIMITIVE_TYPE_MAP[type_annotation.a_type]

    elif isinstance(type_annotation, intermediate.OurTypeAnnotation):
        our_type = type_annotation.our_type

        if isinstance(our_type, intermediate.Enumeration):
            enum_name = golang_naming.enum_name(type_annotation.our_type.name)
            if types_package is None:
                return enum_name

            return Stripped(f"{types_package}.{enum_name}")

        elif isinstance(our_type, intermediate.ConstrainedPrimitive):
            return PRIMITIVE_TYPE_MAP[our_type.constrainee]

        elif isinstance(
            our_type, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            # NOTE (mristin, 2023-03-28):
            # We always refer to interfaces even in cases of concrete classes without
            # concrete descendants since we want to allow enhancing.
            interface_name = golang_naming.interface_name(our_type.name)

            if types_package is None:
                return interface_name

            return Stripped(f"{types_package}.{interface_name}")

    elif isinstance(type_annotation, intermediate.ListTypeAnnotation):
        item_type = generate_type(
            type_annotation=type_annotation.items, types_package=types_package
        )

        return Stripped(f"[]{item_type}")

    elif isinstance(type_annotation, intermediate.OptionalTypeAnnotation):
        value_type = generate_type(
            type_annotation=type_annotation.value, types_package=types_package
        )

        if golang_pointering.is_pointer_type(type_annotation):
            return Stripped(f"*{value_type}")

        return value_type

    else:
        assert_never(type_annotation)

    raise AssertionError("Should not have gotten here")


INDENT2 = INDENT * 2
INDENT3 = INDENT * 3
INDENT4 = INDENT * 4
INDENT5 = INDENT * 5
INDENT6 = INDENT * 6

WARNING = Stripped(
    """\
// This code has been automatically generated by aas-core-codegen.
// Do NOT edit or append."""
)


class GeneratorForLoopVariables:
    """
    Generate a unique variable name based on ``item`` stem.

    >>> generator = GeneratorForLoopVariables()

    >>> next(generator)
    'v'

    >>> next(generator)
    'v1'

    >>> next(generator)
    'v2'
    """

    def __init__(self) -> None:
        """Initialize with the zero counter."""
        self.counter = 0

    def __next__(self) -> Identifier:
        """Generate the next variable name."""
        if self.counter == 0:
            result = Identifier("v")
        else:
            result = Identifier(f"v{self.counter}")

        self.counter += 1

        return result



"""Generate the C++ data structures from the intermediate representation."""
import io
import itertools
import textwrap
from typing import (
    Optional,
    Dict,
    List,
    Tuple,
    cast,
    Union,
)

from icontract import ensure, require

from aas_core_codegen import intermediate
from aas_core_codegen import specific_implementations
from aas_core_codegen.common import (
    Error,
    Identifier,
    assert_never,
    Stripped,
    indent_but_first_line,
)
from aas_core_codegen.cpp import (
    common as cpp_common,
    naming as cpp_naming,
    description as cpp_description,
)
from aas_core_codegen.cpp.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)
from aas_core_codegen.intermediate import construction as intermediate_construction


# region Checks


def _human_readable_identifier(
    something: Union[
        intermediate.Enumeration,
        intermediate.AbstractClass,
        intermediate.ConcreteClass,
        intermediate.EnumerationLiteral,
    ]
) -> str:
    """
    Represent ``something`` in a human-readable text.

    The reader should be able to trace ``something`` back to the meta-model.
    """
    # NOTE (mristin, 2023-03-29):
    # This function has been copy-pasted from
    # :py:mod:`aas_core_codegen.python.structure._generate`. We tried to refactor it to
    # :py:mod:`aas_core_codegen.intermediate`, but it turned out that the refactored
    # code was nigh unreadable. So we preferred a little bit of copying to a little
    # bit of complexity.

    result: str

    if isinstance(something, intermediate.Enumeration):
        result = f"meta-model enumeration {something.name!r}"
    elif isinstance(something, intermediate.EnumerationLiteral):
        result = f"meta-model enumeration literal {something.name!r}"
    elif isinstance(something, intermediate.AbstractClass):
        result = f"meta-model abstract class {something.name!r}"
    elif isinstance(something, intermediate.ConcreteClass):
        result = f"meta-model concrete class {something.name!r}"
    else:
        assert_never(something)

    return result


def _verify_intra_structure_collisions(
    our_type: intermediate.OurType,
) -> Optional[Error]:
    """Verify that no member names collide in the Golang structure of our type."""
    errors = []  # type: List[Error]

    if isinstance(our_type, intermediate.Enumeration):
        observed_literal_names = {}  # type: Dict[Identifier, str]
        for literal in our_type.literals:
            name = cpp_naming.enum_literal_name(literal.name)

            if name in observed_literal_names:
                errors.append(
                    Error(
                        literal.parsed.node,
                        f"C++ name {name!r} corresponding "
                        f"to the meta-model enumeration literal {literal.name!r} "
                        f"collides with the literal {observed_literal_names[name]}",
                    )
                )
            else:
                observed_literal_names[name] = literal.name

    elif isinstance(our_type, intermediate.ConstrainedPrimitive):
        pass

    elif isinstance(our_type, intermediate.Class):
        observed_member_names = {}  # type: Dict[Identifier, str]

        for prop in our_type.properties:
            name = cpp_naming.getter_name(prop.name)
            if name in observed_member_names:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"C++ getter {name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[name]}",
                    )
                )
            else:
                observed_member_names[name] = (
                    f"C++ getter {name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

            name = cpp_naming.mutable_getter_name(prop.name)
            if name in observed_member_names:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"C++ mutable getter {name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[name]}",
                    )
                )
            else:
                observed_member_names[name] = (
                    f"C++ mutable getter {name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

            name = cpp_naming.setter_name(prop.name)
            if name in observed_member_names:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"C++ setter {name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[name]}",
                    )
                )
            else:
                observed_member_names[name] = (
                    f"C++ setter {name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

            name = cpp_naming.private_property_name(prop.name)
            if name in observed_member_names:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"C++ private property {name!r} corresponding "
                        f"to the meta-model property {prop.name!r} collides with "
                        f"the {observed_member_names[name]}",
                    )
                )
            else:
                observed_member_names[name] = (
                    f"C++ private property {name!r} corresponding to "
                    f"the meta-model property {prop.name!r}"
                )

        for method in our_type.methods:
            method_name = cpp_naming.method_name(method.name)

            if method_name in observed_member_names:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"C++ method {method_name!r} corresponding "
                        f"to the meta-model method {method.name!r} collides with "
                        f"the {observed_member_names[method_name]}",
                    )
                )
            else:
                observed_member_names[method_name] = (
                    f"C++ method {method_name!r} corresponding to "
                    f"the meta-model method {method.name!r}"
                )

    else:
        assert_never(our_type)

    if len(errors) > 0:
        errors.append(
            Error(
                our_type.parsed.node,
                f"Naming collision(s) in C++ code for our type {our_type.name!r}",
                underlying=errors,
            )
        )

    return None


def _verify_structure_name_collisions(
    symbol_table: intermediate.SymbolTable,
) -> List[Error]:
    """Verify that the C++ names of the structures do not collide."""
    observed_type_names: Dict[
        Identifier,
        Union[
            intermediate.Enumeration,
            intermediate.AbstractClass,
            intermediate.ConcreteClass,
        ],
    ] = dict()

    errors = []  # type: List[Error]

    # region Inter-structure collisions

    for enum_or_cls in itertools.chain(symbol_table.enumerations, symbol_table.classes):
        names: List[Identifier]

        if isinstance(enum_or_cls, intermediate.Enumeration):
            names = [cpp_naming.enum_name(enum_or_cls.name)]
        elif isinstance(enum_or_cls, intermediate.AbstractClass):
            names = [cpp_naming.interface_name(enum_or_cls.name)]
        elif isinstance(enum_or_cls, intermediate.ConcreteClass):
            names = [
                cpp_naming.interface_name(enum_or_cls.name),
                cpp_naming.class_name(enum_or_cls.name),
            ]
        else:
            assert_never(enum_or_cls)

        for name in names:
            other = observed_type_names.get(name, None)

            if other is not None:
                errors.append(
                    Error(
                        enum_or_cls.parsed.node,
                        f"The C++ name {name!r} "
                        f"of the {_human_readable_identifier(enum_or_cls)} "
                        f"collides with the C++ name "
                        f"of the {_human_readable_identifier(other)}",
                    )
                )
            else:
                observed_type_names[name] = enum_or_cls

    # endregion

    # region Intra-structure collisions

    for our_type in symbol_table.our_types:
        collision_error = _verify_intra_structure_collisions(our_type=our_type)

        if collision_error is not None:
            errors.append(collision_error)

    # endregion

    return errors


class VerifiedIntermediateSymbolTable(intermediate.SymbolTable):
    """Represent a verified symbol table which can be used for code generation."""

    # noinspection PyInitNewSignature
    def __new__(
        cls, symbol_table: intermediate.SymbolTable
    ) -> "VerifiedIntermediateSymbolTable":
        raise AssertionError("Only for type annotation")


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def verify(
    symbol_table: intermediate.SymbolTable,
) -> Tuple[Optional[VerifiedIntermediateSymbolTable], Optional[List[Error]]]:
    """Verify that Golang code can be generated from the ``symbol_table``."""
    errors = []  # type: List[Error]

    structure_name_collisions = _verify_structure_name_collisions(
        symbol_table=symbol_table
    )

    errors.extend(structure_name_collisions)

    if len(errors) > 0:
        return None, errors

    return cast(VerifiedIntermediateSymbolTable, symbol_table), None


# endregion

# region Generation


@require(lambda enumeration, literal: id(literal) in enumeration.literal_id_set)
@require(lambda literal: literal.description is not None)
def _generate_comment_for_enumeration_literal(
    enumeration: intermediate.Enumeration,
    literal: intermediate.EnumerationLiteral,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for the given enumeration literal."""
    # NOTE (mristin, 2023-07-14):
    # We need to state the pre-condition for the second time for mypy.
    assert literal.description is not None

    # fmt: off
    comment, errors = (
        cpp_description.generate_comment_for_summary_remarks(
            description=literal.description,
            context=cpp_description.Context(
                namespace=cpp_common.TYPES_NAMESPACE,
                cls_or_enum=enumeration
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


@require(lambda cls_or_enum: cls_or_enum.description is not None)
def _generate_comment_for_cls_or_enum(
    cls_or_enum: Union[intermediate.Enumeration, intermediate.ClassUnion],
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the documentation comment for our type."""
    # NOTE (mristin, 2023-07-14):
    # We need to state the pre-condition for the second time for mypy.
    assert cls_or_enum.description is not None

    # fmt: off
    comment, errors = (
        cpp_description
        .generate_comment_for_summary_remarks_constraints(
            description=cls_or_enum.description,
            context=cpp_description.Context(
                namespace=cpp_common.TYPES_NAMESPACE,
                cls_or_enum=None
            )
        )
    )
    # fmt: on

    if errors is not None:
        return None, errors

    assert comment is not None

    return comment, None


@ensure(lambda result: (result[0] is None) ^ (result[1] is None))
def _generate_enum(
    enum: intermediate.Enumeration,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the C++ code for the enum."""
    writer = io.StringIO()

    errors = []  # type: List[Error]

    comment = None  # type: Optional[Stripped]
    if enum.description is not None:
        comment, comment_errors = _generate_comment_for_cls_or_enum(cls_or_enum=enum)

        if comment_errors:
            errors.append(
                Error(
                    enum.description.parsed.node,
                    f"Failed to generate the comment "
                    f"for the enumeration {enum.name!r}",
                    comment_errors,
                )
            )
        else:
            assert comment is not None

    if comment is not None:
        writer.write(comment)
        writer.write("\n")

    name = cpp_naming.enum_name(enum.name)

    writer.write(f"enum class {name} : std::uint32_t {{\n")
    if len(enum.literals) == 0:
        writer.write("// Intentionally empty.")
    else:
        for i, literal in enumerate(enum.literals):
            if literal.description is not None:
                comment, comment_errors = _generate_comment_for_enumeration_literal(
                    enumeration=enum, literal=literal
                )
                if comment_errors is not None:
                    errors.append(
                        Error(
                            literal.description.parsed.node,
                            f"Failed to generate the documentation comment "
                            f"for enumeration literal {literal.name!r}",
                            comment_errors,
                        )
                    )
                else:
                    assert comment is not None
                    writer.write(textwrap.indent(comment, I))
                    writer.write("\n")

            literal_name = cpp_naming.enum_literal_name(literal.name)

            writer.write(textwrap.indent(f"{literal_name} = {i}", I))
            if i < len(enum.literals):
                writer.write(",\n")
            else:
                writer.write("\n")

        writer.write(f"}};  // enum class {name}")

    if len(errors) > 0:
        return None, Error(
            enum.parsed.node,
            f"Failed to generate the Golang code for the enumeration {enum.name!r}",
            errors,
        )

    return Stripped(writer.getvalue()), None


def _generate_literals_of_enum_definition(enum: intermediate.Enumeration) -> Stripped:
    """Generate the definition for the constant vector listing the literals."""
    enum_name = cpp_naming.enum_name(enum.name)
    constant_name = cpp_naming.constant_name(Identifier(f"literals_of_{enum.name}"))
    return Stripped(
        f"""\
extern const std::vector<
{I}{enum_name}
> {constant_name};"""
    )


def _generate_literals_of_enum_implementation(
    enum: intermediate.Enumeration,
) -> Stripped:
    """Generate the implementation for the constant vector listing the literals."""
    enum_name = cpp_naming.enum_name(enum.name)
    constant_name = cpp_naming.constant_name(Identifier(f"literals_of_{enum.name}"))

    literals = []
    for literal in enum.literals:
        literal_name = cpp_naming.enum_literal_name(literal.name)
        literals.append(f"{enum_name}::{literal_name}")

    literals_joined = ",\n".join(literals)

    return Stripped(
        f"""\
/**
 * List the literals of {enum_name}.
 *
 * C++ does not provide an elegant way to iterate over the literals, so
 * this array helps you avoid common errors and pitfalls.
 */
const std::vector<
{I}{enum_name}
> {constant_name} = {{
{I}{indent_but_first_line(literals_joined, I)}
}};"""
    )


def _generate_model_type_definition(
    symbol_table: intermediate.SymbolTable,
) -> Stripped:
    """
    Generate the enumeration corresponding to the model types.

    This is necessary so that we can have fast switches on instance types.
    """
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))

    literal_specs = []  # type: List[Stripped]
    for i, cls in enumerate(symbol_table.concrete_classes):
        literal_name = cpp_naming.enum_literal_name(cls.name)

        literal_specs.append(Stripped(f"{literal_name} = {i}"))

    literal_specs_joined = ",\n".join(literal_specs)

    if len(literal_specs) == 0:
        return Stripped(
            f"""\
enum class {enum_name} : std::uint32_t {{
{I}// Intentionally empty.
}};"""
        )

    return Stripped(
        f"""\
/**
 * Enumerate the model types for faster type switches.
 *
 * For example, switch statements can be implemented as jump tables.
 */
enum class {enum_name} : std::uint32_t {{
{I}{indent_but_first_line(literal_specs_joined, I)}
}};"""
    )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_class_interface(
    cls: intermediate.ClassUnion,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition of an interface representing a class."""
    errors = []  # type: List[Error]

    members = []  # type: List[Stripped]

    for prop in cls.properties:
        if prop.specified_for is not cls:
            continue

        description_comment = None  # type: Optional[Stripped]
        if prop.description is not None:
            (
                description_comment,
                description_errors,
            ) = cpp_description.generate_comment_for_summary_remarks(
                prop.description,
                context=cpp_description.Context(
                    namespace=cpp_common.TYPES_NAMESPACE, cls_or_enum=cls
                ),
            )
            if description_errors is not None:
                errors.append(
                    Error(
                        prop.parsed.node,
                        f"Failed to generate the description comment "
                        f"for the property {prop.name!r}",
                        description_errors,
                    )
                )
            else:
                assert description_comment is not None
                members.append(
                    Stripped(
                        f"""\
///@{{
{description_comment}"""
                    )
                )

        getter_name = cpp_naming.getter_name(prop.name)
        getter_type = cpp_common.generate_type_with_const_ref_if_applicable(
            prop.type_annotation
        )

        members.append(Stripped(f"virtual {getter_type} {getter_name}() const = 0;"))

        mutable_getter_name = cpp_naming.mutable_getter_name(prop.name)
        mutable_getter_type = cpp_common.generate_type_with_ref(prop.type_annotation)
        members.append(
            Stripped(f"virtual {mutable_getter_type} {mutable_getter_name}() = 0;")
        )

        setter_name = cpp_naming.setter_name(prop.name)
        # NOTE (mristin, 2023-09-13):
        # For a discussion on ``std::shared_ptr`` and referencing,
        # see: https://herbsutter.com/2013/06/05/gotw-91-solution-smart-pointer-parameters/

        # NOTE (mristin, 2023-09-13):
        # We provide setters which only set-by-value logic since this seems to be
        # the best approach for general cases,
        # see: https://stackoverflow.com/questions/10692345/is-it-worth-adding-a-move-enabled-setter.
        #
        # Whenever you know that you do not need the value after calling the setter,
        # make sure to call it with ``std::move(.)``.

        value_type = cpp_common.generate_type(type_annotation=prop.type_annotation)
        members.append(
            Stripped(
                f"""\
virtual void {setter_name}(
{I}{indent_but_first_line(value_type, I)} value
) = 0;"""
            )
        )

        if description_comment is not None:
            members.append(Stripped("///@}"))

    for method in cls.methods:
        if method.specified_for is not cls:
            continue

        returns = (
            cpp_common.generate_type(method.returns)
            if method.returns is not None
            else None
        )

        return_type = "void" if returns is None else returns

        arg_types_names = [
            (
                cpp_common.generate_type_with_const_ref_if_applicable(
                    arg.type_annotation
                ),
                cpp_naming.argument_name(arg.name),
            )
            for arg in method.arguments
        ]

        method_name = cpp_naming.method_name(method.name)

        const_suffix = " const" if method.non_mutating else ""

        description_comment_prefix = ""
        if method.description is not None:
            (
                description_comment,
                description_errors,
            ) = cpp_description.generate_comment_for_summary_remarks(
                method.description,
                context=cpp_description.Context(
                    namespace=cpp_common.TYPES_NAMESPACE, cls_or_enum=cls
                ),
            )
            if description_errors is not None:
                errors.append(
                    Error(
                        method.parsed.node,
                        f"Failed to generate the description comment "
                        f"for the method {method.name!r}",
                        description_errors,
                    )
                )
            else:
                assert description_comment is not None
                description_comment_prefix = description_comment + "\n"

        if len(method.arguments) == 0:
            members.append(
                Stripped(
                    f"{description_comment_prefix}"
                    f"virtual {return_type} {method_name}(){const_suffix} = 0;"
                )
            )
        else:
            arguments_definition = ",\n".join(
                f"{arg_type} {arg_name}" for arg_type, arg_name in arg_types_names
            )

            members.append(
                Stripped(
                    f"""\
{description_comment_prefix}virtual {return_type} {method_name}(
{I}{indent_but_first_line(arguments_definition, I)}
){const_suffix} = 0;"""
                )
            )

    interface_name = cpp_naming.interface_name(cls.name)

    members.append(Stripped(f"virtual ~{interface_name}() = default;"))

    members_joined = "\n\n".join(members)

    inheritances = [
        f"virtual public {cpp_naming.interface_name(inheritance.name)}"
        for inheritance in cls.inheritances
    ]

    if len(inheritances) == 0:
        inheritances = ["virtual public IClass"]

    inheritances_joined = ",\n".join(inheritances)

    description_comment = None
    if cls.description is not None:
        (
            description_comment,
            description_errors,
        ) = cpp_description.generate_comment_for_summary_remarks_constraints(
            description=cls.description,
            context=cpp_description.Context(
                namespace=cpp_common.TYPES_NAMESPACE, cls_or_enum=cls
            ),
        )
        if description_errors is not None:
            errors.append(
                Error(
                    cls.description.parsed.node,
                    "Failed to translate the class description to a comment",
                    description_errors,
                )
            )
        else:
            assert description_comment is not None

    if len(errors) > 0:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate the class definition for {cls.name!r}",
            errors,
        )

    maybe_description_comment_nl = (
        "" if description_comment is None else description_comment + "\n"
    )

    return (
        Stripped(
            f"""\
{maybe_description_comment_nl}class {interface_name}
{II}: {indent_but_first_line(inheritances_joined, II)} {{
 public:
{I}{indent_but_first_line(members_joined, I)}
}};"""
        ),
        None,
    )


def _generate_class_definition(cls: intermediate.ConcreteClass) -> Stripped:
    """Generate the definition of a concrete class."""
    public_members = []  # type: List[Stripped]

    cls_name = cpp_naming.class_name(cls.name)

    if len(cls.constructor.arguments) == 0:
        public_members.append(Stripped(f"{cls_name}() {{}}"))
    else:
        constructor_argument_specs = []  # type: List[str]
        for arg in cls.constructor.arguments:
            arg_type = cpp_common.generate_type(arg.type_annotation)
            arg_name = cpp_naming.argument_name(arg.name)

            assign_default_suffix = (
                " = common::nullopt"
                if isinstance(arg.type_annotation, intermediate.OptionalTypeAnnotation)
                else ""
            )

            constructor_argument_specs.append(
                f"{arg_type} {arg_name}{assign_default_suffix}"
            )

        constructor_arguments_specs_joined = ",\n".join(constructor_argument_specs)

        # NOTE (mristin, 2023-09-22):
        # See: https://stackoverflow.com/questions/12437241/c-always-use-explicit-constructor
        non_default_argument_count = sum(
            1 for arg in cls.constructor.arguments if arg.default is None
        )

        explicit_prefix = "explicit " if non_default_argument_count == 1 else ""

        public_members.append(
            Stripped(
                f"""\
{explicit_prefix}{cls_name}(
{I}{indent_but_first_line(constructor_arguments_specs_joined, I)}
);"""
            )
        )

    model_type_getter = cpp_naming.getter_name(Identifier("model_type"))
    model_type_enum = cpp_naming.enum_name(Identifier("Model_type"))

    public_members.append(
        Stripped(f"{model_type_enum} {model_type_getter}() const override;")
    )

    for prop in cls.properties:
        getter_name = cpp_naming.getter_name(prop.name)
        getter_type = cpp_common.generate_type_with_const_ref_if_applicable(
            prop.type_annotation
        )

        public_members.append(
            Stripped(
                f"// region Get and set {cpp_naming.private_property_name(prop.name)}"
            )
        )

        public_members.append(
            Stripped(f"{getter_type} {getter_name}() const override;")
        )

        mutable_getter_name = cpp_naming.mutable_getter_name(prop.name)
        mutable_getter_type = cpp_common.generate_type_with_ref(prop.type_annotation)
        public_members.append(
            Stripped(f"{mutable_getter_type} {mutable_getter_name}() override;")
        )

        setter_name = cpp_naming.setter_name(prop.name)

        value_type = cpp_common.generate_type(type_annotation=prop.type_annotation)
        public_members.append(
            Stripped(
                f"""\
void {setter_name}(
{I}{indent_but_first_line(value_type, I)} value
) override;"""
            )
        )

        public_members.append(Stripped("// endregion"))

    for method in cls.methods:
        returns = (
            cpp_common.generate_type(method.returns)
            if method.returns is not None
            else None
        )

        return_type = "void" if returns is None else returns

        arg_types_names = [
            (
                cpp_common.generate_type_with_const_ref_if_applicable(
                    arg.type_annotation
                ),
                cpp_naming.argument_name(arg.name),
            )
            for arg in method.arguments
        ]

        method_name = cpp_naming.method_name(method.name)

        const_suffix = " const" if method.non_mutating else ""

        if len(method.arguments) == 0:
            public_members.append(
                Stripped(f"{return_type} {method_name}(){const_suffix} override;")
            )
        else:
            arguments_definition = ",\n".join(
                f"{arg_type} {arg_name}" for arg_type, arg_name in arg_types_names
            )

            public_members.append(
                Stripped(
                    f"""\
{return_type} {method_name}(
{I}{indent_but_first_line(arguments_definition, I)}
){const_suffix} override;"""
                )
            )

    public_members.append(Stripped(f"~{cls_name}() override = default;"))

    public_members_joined = "\n\n".join(public_members)

    interface_name = cpp_naming.interface_name(cls.name)

    private_members = []  # type: List[Stripped]
    for prop in cls.properties:
        value_type = cpp_common.generate_type(type_annotation=prop.type_annotation)

        member_name = cpp_naming.private_property_name(prop.name)
        private_members.append(Stripped(f"{value_type} {member_name};"))

    blocks = [
        f"""\
 public:
{I}{indent_but_first_line(public_members_joined, I)}"""
    ]

    if len(private_members) > 0:
        private_members_joined = "\n\n".join(private_members)
        blocks.append(
            f"""\
 private:
{I}{indent_but_first_line(private_members_joined, I)}"""
        )

    blocks_joined = "\n\n".join(blocks)

    return Stripped(
        f"""\
class {cls_name}
{II}: public {interface_name} {{
{blocks_joined}
}};"""
    )


def _generate_is_cls_definition(cls: intermediate.ClassUnion) -> Stripped:
    """Generate the definition of the function to check is-a based on model type."""
    function_name = cpp_naming.function_name(Identifier(f"is_{cls.name}"))

    interface_name = cpp_naming.interface_name(cls.name)

    return Stripped(
        f"""\
/**
 * \\brief Check whether \\p that instance is of runtime type
 * \\ref {interface_name}.
 *
 * We use `IClass::model_type` to determine the runtime type, which is
 * a bit faster than native C++'s RTTI.
 *
 * \\param that instance to check for runtime type
 * \\return `true` if \\p that instance is indeed
 * an instance of \\ref {interface_name}
 */
bool {function_name}(
{I}const IClass& that
);"""
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_header(
    symbol_table: VerifiedIntermediateSymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
    library_namespace: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the C++ header code of the structures based on the symbol table."""
    namespace = Stripped(f"{library_namespace}::{cpp_common.TYPES_NAMESPACE}")

    include_guard_var = cpp_common.include_guard_var(namespace)

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        Stripped(
            f"""\
#ifndef {include_guard_var}
#define {include_guard_var}"""
        ),
        cpp_common.WARNING,
        Stripped(
            f"""\
#include "{include_prefix_path}/common.hpp"

#pragma warning(push, 0)
#include <cstddef>
#include <cstdint>
#include <memory>
#include <string>
#include <vector>
#pragma warning(pop)"""
        ),
        cpp_common.generate_namespace_opening(library_namespace),
        Stripped(
            f"""\
/**
 * \\defgroup types Define data structures corresponding to the meta-model.
 * @{{
 */
namespace {cpp_common.TYPES_NAMESPACE} {{"""
        ),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    blocks.append(Stripped("// region Enumerations"))

    blocks.append(_generate_model_type_definition(symbol_table=symbol_table))

    for enum in symbol_table.enumerations:
        block, error = _generate_enum(enum=enum)
        if error is not None:
            errors.append(error)
        else:
            assert block is not None
            blocks.append(block)

    blocks.append(Stripped("// endregion Enumerations"))

    blocks.append(Stripped("// region Forward declaration of interfaces"))

    blocks.append(Stripped("// endregion Forward declaration of interfaces"))

    for cls in symbol_table.classes:
        interface_name = cpp_naming.interface_name(cls.name)
        blocks.append(Stripped(f"class {interface_name};"))

    blocks.append(Stripped("// region Class interfaces"))

    model_type_getter = cpp_naming.getter_name(Identifier("model_type"))
    model_type_enum = cpp_naming.enum_name(Identifier("Model_type"))

    blocks.append(
        Stripped(
            f"""\
/**
 * Model the most general instance of the model.
 */
class IClass {{
 public:
{I}/**
{I} * Indicate the runtime model type.
{I} */
{I}virtual {model_type_enum} {model_type_getter}() const = 0;
{I}virtual ~IClass() = default;
}};"""
        )
    )

    for cls in symbol_table.classes:
        block, error = _generate_class_interface(cls=cls)
        if error is not None:
            errors.append(error)
        else:
            assert block is not None
            blocks.append(block)

    blocks.append(Stripped("// endregion"))

    blocks.append(Stripped("// region Definitions of concrete classes"))

    for cls in symbol_table.concrete_classes:
        blocks.append(_generate_class_definition(cls=cls))

    blocks.append(Stripped("// endregion"))

    if len(errors) > 0:
        return None, errors

    blocks.append(Stripped("// region Is-a functions"))

    for cls in symbol_table.classes:
        blocks.append(_generate_is_cls_definition(cls=cls))

    blocks.append(Stripped("// endregion Is-a functions"))

    blocks.extend(
        [
            Stripped(
                f"""\
}}  // namespace {cpp_common.TYPES_NAMESPACE}
/**@}}*/"""
            ),
            cpp_common.generate_namespace_closing(library_namespace),
            cpp_common.WARNING,
            Stripped(f"#endif  // {include_guard_var}"),
        ]
    )

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        out.write(block)

    out.write("\n")

    return out.getvalue(), None


def _generate_constructor_implementation(cls: intermediate.ConcreteClass) -> Stripped:
    """Transpile the constructor implementation for the given ``cls``."""
    assert all(
        isinstance(stmt, intermediate_construction.AssignArgument)
        for stmt in cls.constructor.inlined_statements
    ), (
        f"We expect only assigns in the inlined constructors, "
        f"but got for class {cls.name!r}: {cls.constructor.inlined_statements}"
    )

    body_statements = []  # type: List[Stripped]

    for stmt in cls.constructor.inlined_statements:
        assert isinstance(stmt, intermediate_construction.AssignArgument), (
            f"Only assigns expected in inlined constructors, but got the following "
            f"statement for the class {cls.name!r}: {stmt}"
        )

        prop_member = cpp_naming.private_property_name(stmt.name)
        arg_name = cpp_naming.argument_name(stmt.argument)

        if stmt.default is not None:
            if isinstance(stmt.default, intermediate_construction.EmptyList):
                body_statements.append(
                    Stripped(
                        f"""\
if ({arg_name}.has_value()) {{
{I}{prop_member} = *{arg_name};
}} else {{
{I}{prop_member}.emplace();
}}"""
                    )
                )
            elif isinstance(stmt.default, intermediate_construction.DefaultEnumLiteral):
                enum_name = cpp_naming.enum_name(stmt.default.enum.name)
                literal_name = cpp_naming.enum_literal_name(stmt.default.literal.name)

                body_statements.append(
                    Stripped(
                        f"""\
{prop_member} = ({arg_name}.has_value())
{I}? *{arg_name}
{I}: {enum_name}::{literal_name};"""
                    )
                )
            else:
                assert_never(stmt.default)

        else:
            prop = cls.properties_by_name[stmt.name]
            if cpp_common.is_referencable(prop.type_annotation):
                body_statements.append(
                    Stripped(f"{prop_member} = std::move({arg_name});")
                )
            else:
                body_statements.append(Stripped(f"{prop_member} = {arg_name};"))

    body = (
        "\n\n".join(body_statements)
        if len(body_statements) > 0
        else "// Intentionally empty."
    )

    cls_name = cpp_naming.class_name(cls.name)

    if len(cls.constructor.arguments) == 0:
        return Stripped(
            f"""\
{cls_name}::{cls_name}() {{
{I}{indent_but_first_line(body, I)}
}}"""
        )

    constructor_argument_specs = []  # type: List[str]
    for arg in cls.constructor.arguments:
        arg_type = cpp_common.generate_type(arg.type_annotation)
        arg_name = cpp_naming.argument_name(arg.name)

        constructor_argument_specs.append(f"{arg_type} {arg_name}")

    constructor_arguments_specs_joined = ",\n".join(constructor_argument_specs)

    return Stripped(
        f"""\
{cls_name}::{cls_name}(
{I}{indent_but_first_line(constructor_arguments_specs_joined, I)}
) {{
{I}{indent_but_first_line(body, I)}
}}"""
    )


def _generate_model_type_getter_implementation(
    cls: intermediate.ConcreteClass,
) -> Stripped:
    """Implement the getter for the runtime model type."""
    enum_name = cpp_naming.enum_name(Identifier("Model_type"))
    literal_name = cpp_naming.enum_literal_name(cls.name)

    cls_name = cpp_naming.class_name(cls.name)
    getter_name = cpp_naming.getter_name(Identifier("model_type"))

    return Stripped(
        f"""\
{enum_name} {cls_name}::{getter_name}() const {{
{I}return {enum_name}::{literal_name};
}}"""
    )


@require(lambda prop, cls: id(prop) in cls.property_id_set)
def _generate_getters_and_setter(
    prop: intermediate.Property, cls: intermediate.ConcreteClass
) -> List[Stripped]:
    """Generate the immutable and mutable getter and the setter."""
    cls_name = cpp_naming.class_name(cls.name)
    member_name = cpp_naming.private_property_name(prop.name)

    blocks = []  # type: List[Stripped]

    # region Getter
    getter_name = cpp_naming.getter_name(prop.name)
    getter_type = cpp_common.generate_type_with_const_ref_if_applicable(
        prop.type_annotation
    )

    blocks.append(
        Stripped(
            f"""\
{getter_type} {cls_name}::{getter_name}() const {{
{I}return {member_name};
}}"""
        )
    )
    # endregion

    # region Mutable getter
    mutable_getter_name = cpp_naming.mutable_getter_name(prop.name)
    mutable_getter_type = cpp_common.generate_type_with_ref(prop.type_annotation)

    blocks.append(
        Stripped(
            f"""\
{mutable_getter_type} {cls_name}::{mutable_getter_name}() {{
{I}return {member_name};
}}"""
        )
    )
    # endregion

    # region Setter
    setter_name = cpp_naming.setter_name(prop.name)
    value_type = cpp_common.generate_type(type_annotation=prop.type_annotation)

    blocks.append(
        Stripped(
            f"""\
void {cls_name}::{setter_name}(
{I}{indent_but_first_line(value_type, I)} value
) {{
{I}{member_name} = value;
}}"""
        )
    )
    # endregion

    return blocks


@require(lambda method, cls: id(method) in cls.method_id_set)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_method_implementation(
    method: intermediate.MethodUnion,
    cls: intermediate.ConcreteClass,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the implementation of the method."""
    body: Optional[Stripped]

    if isinstance(method, intermediate.ImplementationSpecificMethod):
        implementation_key = specific_implementations.ImplementationKey(
            f"types/{method.specified_for.name}/{method.name}.body.cpp"
        )

        body = spec_impls.get(implementation_key, None)

        if body is None:
            return None, Error(
                method.parsed.node,
                f"The implementation is missing for "
                f"the implementation-specific method: {implementation_key}",
            )

    elif isinstance(method, intermediate.UnderstoodMethod):
        return None, Error(
            cls.parsed.node,
            "At the moment (2023-09-22), we do not transpile the method body and "
            "its contracts, as it is quite a difficult task. Please contact "
            "the developers if you need this feature.",
        )
    else:
        assert_never(method)

    assert body is not None

    returns = (
        cpp_common.generate_type(method.returns) if method.returns is not None else None
    )

    return_type = "void" if returns is None else returns

    arg_types_names = [
        (
            cpp_common.generate_type_with_const_ref_if_applicable(arg.type_annotation),
            cpp_naming.argument_name(arg.name),
        )
        for arg in method.arguments
    ]

    method_name = cpp_naming.method_name(method.name)

    const_suffix = " const" if method.non_mutating else ""

    cls_name = cpp_naming.class_name(cls.name)

    if len(method.arguments) == 0:
        return (
            Stripped(
                f"""\
{return_type} {cls_name}::{method_name}(){const_suffix} {{
{I}{indent_but_first_line(body, I)}
}}"""
            ),
            None,
        )
    else:
        arguments_definition = ",\n".join(
            f"{arg_type} {arg_name}" for arg_type, arg_name in arg_types_names
        )

        return (
            Stripped(
                f"""\
{return_type} {cls_name}::{method_name}(
{I}{indent_but_first_line(arguments_definition, I)}
){const_suffix} {{
{I}{indent_but_first_line(body, I)}
}}"""
            ),
            None,
        )


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _generate_class_implementation(
    cls: intermediate.ConcreteClass,
    spec_impls: specific_implementations.SpecificImplementations,
) -> Tuple[Optional[List[Stripped]], Optional[Error]]:
    """Generate the implementation blocks for the given class."""
    blocks = [
        _generate_constructor_implementation(cls=cls),
        _generate_model_type_getter_implementation(cls=cls),
    ]  # type: List[Stripped]

    for prop in cls.properties:
        blocks.extend(_generate_getters_and_setter(prop=prop, cls=cls))

    errors = []  # type: List[Error]

    for method in cls.methods:
        code, error = _generate_method_implementation(
            method=method, cls=cls, spec_impls=spec_impls
        )
        if error is not None:
            errors.append(
                Error(
                    method.parsed.node,
                    f"Failed to generate the implementation "
                    f"for the method {method.name!r}",
                    [error],
                )
            )
        else:
            assert code is not None
            blocks.append(code)

    if len(errors) > 0:
        return None, Error(
            cls.parsed.node,
            f"Failed to generate the implementation for the class {cls.name!r}",
            errors,
        )

    return blocks, None


def _generate_is_cls_implementation(
    cls: intermediate.ClassUnion, symbol_table: intermediate.SymbolTable
) -> Stripped:
    """Generate the impl. of the function to check is-a based on model type."""
    function_name = cpp_naming.function_name(Identifier(f"is_{cls.name}"))

    case_blocks = []  # type: List[Stripped]
    for concrete_cls in symbol_table.concrete_classes:
        case_body = (
            "return true;" if concrete_cls.is_subclass_of(cls) else "return false;"
        )

        model_type_literal = cpp_naming.enum_literal_name(concrete_cls.name)

        case_blocks.append(
            Stripped(
                f"""\
case ModelType::{model_type_literal}:
{I}{indent_but_first_line(case_body, I)}"""
            )
        )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::invalid_argument(
{II}common::Concat(
{III}"Unexpected model type: ",
{III}std::to_string(
{IIII}static_cast<std::uint32_t>(that.model_type())
{III})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n".join(case_blocks)

    return Stripped(
        f"""\
bool {function_name}(
{I}const IClass& that
) {{
{I}switch (that.model_type()) {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}}
}}"""
    )


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate_implementation(
    symbol_table: intermediate.SymbolTable,
    spec_impls: specific_implementations.SpecificImplementations,
    library_namespace: Stripped,
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """Generate the C++ implementation code for data structure."""
    namespace = Stripped(f"{library_namespace}::types")

    include_prefix_path = cpp_common.generate_include_prefix_path(library_namespace)

    blocks = [
        cpp_common.WARNING,
        Stripped(
            f'''\
#include "{include_prefix_path}/types.hpp"'''
        ),
        cpp_common.generate_namespace_opening(namespace),
    ]  # type: List[Stripped]

    errors = []  # type: List[Error]

    for concrete_cls in symbol_table.concrete_classes:
        if concrete_cls.is_implementation_specific:
            errors.append(
                Error(
                    concrete_cls.parsed.node,
                    f"We currently do not support implementation-specific classes "
                    f"in the C++ generator, but the class {concrete_cls.name!r} has "
                    f"been marked as implementation-specific. If you need "
                    f"this feature, please contact the developers.",
                )
            )
            continue

        cls_blocks, error = _generate_class_implementation(
            cls=concrete_cls, spec_impls=spec_impls
        )
        if error is not None:
            errors.append(error)
        else:
            assert cls_blocks is not None
            cls_name = cpp_naming.class_name(concrete_cls.name)
            blocks.append(Stripped(f"// region {cls_name}"))
            blocks.extend(cls_blocks)
            blocks.append(Stripped(f"// endregion {cls_name}"))

    if len(errors) > 0:
        return None, errors

    blocks.append(Stripped("// region Is-a functions"))

    for cls in symbol_table.classes:
        blocks.append(
            _generate_is_cls_implementation(cls=cls, symbol_table=symbol_table)
        )

    blocks.append(Stripped("// endregion Is-a functions"))

    blocks.extend(
        [
            cpp_common.generate_namespace_closing(namespace),
            cpp_common.WARNING,
        ]
    )

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(block)

    writer.write("\n")

    return writer.getvalue(), None


# endregion



"""Generate the C++ code for linearized control flows."""
import io
from typing import Sequence, List, Optional

from aas_core_codegen.common import (
    Identifier,
    Stripped,
    pairwise,
    indent_but_first_line,
    assert_never,
)
from aas_core_codegen.yielding import flow as yielding_flow, linear as yielding_linear
from aas_core_codegen.cpp.common import INDENT as I, INDENT2 as II, INDENT3 as III
from aas_core_codegen.cpp import common as cpp_common


def _generate_subroutine_body(
    subroutine: yielding_linear.Subroutine,
    next_subroutine: Optional[yielding_linear.Subroutine],
    state_member: Identifier,
) -> Stripped:
    """Generate the body of the case block for the given subroutine."""
    blocks = []  # type: List[Stripped]

    for i, statement in enumerate(subroutine):
        if isinstance(statement, yielding_linear.Command):
            blocks.append(statement.code)

        elif isinstance(statement, yielding_linear.If):
            if statement.on_true is not None and statement.on_false is not None:
                condition = statement.condition
            elif statement.on_true is not None and statement.on_false is None:
                condition = statement.condition
            elif statement.on_true is None and statement.on_false is not None:
                condition = Stripped(f"!({statement.condition})")
            else:
                raise AssertionError("Unexpected if-statement without target")

            writer = io.StringIO()
            if "\n" in statement.condition or len(statement.condition) > 50:
                writer.write(
                    f"""\
if (
{I}{indent_but_first_line(condition, I)}
) {{
"""
                )
            else:
                writer.write(f"if ({condition}) {{\n")

            if statement.on_true is not None and statement.on_false is not None:
                writer.write(
                    f"""\
{I}{state_member} = {statement.on_true};
}} else {{
{I}{state_member} = {statement.on_false};
}}
continue;"""
                )
            elif statement.on_true is not None and statement.on_false is None:
                writer.write(
                    f"""\
{I}{state_member} = {statement.on_true};
{I}continue;
}}"""
                )
            elif statement.on_true is None and statement.on_false is not None:
                writer.write(
                    f"""\
{I}{state_member} = {statement.on_false};
{I}continue;
}}"""
                )
            else:
                raise AssertionError("Unexpected if-statement without target")

            blocks.append(Stripped(writer.getvalue()))

        elif isinstance(statement, yielding_linear.Jump):
            blocks.append(
                Stripped(
                    f"""\
{state_member} = {statement.target};
continue;"""
                )
            )

        elif isinstance(statement, yielding_linear.Yield):
            if next_subroutine is None:
                assert subroutine[0].label is not None

                blocks.append(
                    Stripped(
                        f"""\
{state_member} = {subroutine[0].label + 1};  // Invalidate state
return;"""
                    )
                )
            else:
                blocks.append(
                    Stripped(
                        f"""\
{state_member} = {next_subroutine[0].label};
return;"""
                    )
                )
        elif isinstance(statement, yielding_linear.Noop):
            if statement.comment is None:
                blocks.append(Stripped("// Noop"))
            else:
                blocks.append(
                    Stripped(cpp_common.non_documentation_comment(statement.comment))
                )
        else:
            assert_never(statement)

        if (
            i == len(subroutine) - 1
            and next_subroutine is None
            and isinstance(statement, yielding_linear.Command)
        ):
            assert subroutine[0].label is not None

            blocks.append(
                Stripped(
                    f"""\
// We invalidate the state since we reached the end of the routine.
{state_member} = {subroutine[0].label + 1};
return;"""
                )
            )

    return Stripped("\n\n".join(blocks))


def generate_execute_body(
    flow: Sequence[yielding_flow.Node], state_member: Identifier
) -> Stripped:
    """
    Generate the code of the ``Execute()`` method which executes the flow with yields.

    The member that holds the state to be executed is indicated by ``state_member``.
    Both the state member and the ``Execute()`` method are usually private.
    """
    if len(flow) == 0:
        return Stripped("// Intentionally empty.")

    subroutines = yielding_linear.linearize_to_subroutines(flow=flow)

    case_blocks = []  # type: List[Stripped]
    for subroutine, next_subroutine in pairwise(subroutines):
        case_body = _generate_subroutine_body(
            subroutine=subroutine,
            next_subroutine=next_subroutine,
            state_member=state_member,
        )
        case_blocks.append(
            Stripped(
                f"""\
case {subroutine[0].label}: {{
{I}{indent_but_first_line(case_body, I)}
}}"""
            )
        )

    last_subroutine = subroutines[-1]
    case_body = _generate_subroutine_body(
        subroutine=last_subroutine, next_subroutine=None, state_member=state_member
    )
    case_blocks.append(
        Stripped(
            f"""\
case {last_subroutine[0].label}: {{
{I}{indent_but_first_line(case_body, I)}
}}"""
        )
    )

    case_blocks.append(
        Stripped(
            f"""\
default:
{I}throw std::logic_error(
{II}common::Concat(
{III}"Invalid {state_member}: ",
{III}std::to_string({state_member})
{II})
{I});"""
        )
    )

    case_blocks_joined = "\n\n".join(case_blocks)

    return Stripped(
        f"""\
while (true) {{
{I}switch ({state_member}) {{
{II}{indent_but_first_line(case_blocks_joined, II)}
{I}}}
}}"""
    )
